from __future__ import annotations

import dataclasses
from typing import Any, Callable, TypeVar, overload

from .._singleton import MISSING_NONPROP

T = TypeVar("T")


@dataclasses.dataclass(frozen=True)
class _SubcommandConfig:
    name: str | None
    default: Any
    description: str | None
    prefix_name: bool
    constructor_factory: Callable[[], type | Callable[..., Any]] | None

    def __hash__(self) -> int:
        return object.__hash__(self)


@overload
def subcommand(
    name: str | None = None,
    *,
    default: Any = MISSING_NONPROP,
    description: str | None = None,
    prefix_name: bool = True,
    constructor: None = None,
    constructor_factory: Callable[[], type | Callable[..., Any]] | None = None,
) -> object: ...


@overload
def subcommand(
    name: str | None = None,
    *,
    default: Any = MISSING_NONPROP,
    description: str | None = None,
    prefix_name: bool = True,
    constructor: type | Callable[..., Any] | None = None,
    constructor_factory: None = None,
) -> object: ...


def subcommand(
    name: str | None = None,
    *,
    default: Any = MISSING_NONPROP,
    description: str | None = None,
    prefix_name: bool = True,
    constructor: type | Callable[..., Any] | None = None,
    constructor_factory: Callable[[], type | Callable[..., Any]] | None = None,
) -> object:
    """Configure subcommand behavior for Union types in the CLI.

    When tyro encounters a Union type over structures, it creates subcommands in the
    CLI. The `subcommand()` function allows you to customize the appearance and behavior
    of these subcommands.

    Example::

        from dataclasses import dataclass
        from typing import Annotated, Union
        import tyro

        @dataclass
        class TrainConfig:
            learning_rate: float = 0.01

        @dataclass
        class EvalConfig:
            checkpoint_path: str

        @dataclass
        class MainConfig:
            # Customized subcommands:
            mode: Union[
                Annotated[TrainConfig, tyro.conf.subcommand("train")],
                Annotated[EvalConfig, tyro.conf.subcommand("evaluate")]
            ]

        # CLI usage: python script.py mode:train --mode.learning-rate 0.02

    If a default value is provided both via `subcommand(default=...)` and in the field
    definition itself (`field = default`), the field default will take precedence.

    Args:
        name: Custom name for the subcommand in the CLI.
        default: Default instance to use for this subcommand.
        description: Custom helptext for this subcommand.
        prefix_name: Whether to include the parent field name as a prefix in the subcommand
            name (default: True).
        constructor: Custom constructor type or function for parsing arguments.
        constructor_factory: Function that returns a constructor type for parsing arguments
            (cannot be used with constructor).

    Returns:
        A configuration object that should be attached to a type using `Annotated[]`.
    """
    assert not (constructor is not None and constructor_factory is not None), (
        "`constructor` and `constructor_factory` cannot both be set."
    )
    return _SubcommandConfig(
        name,
        default,
        description,
        prefix_name,
        constructor_factory=(
            constructor_factory if constructor is None else lambda: constructor
        ),
    )


@dataclasses.dataclass(frozen=True)
class _ArgConfig:
    name: str | None
    metavar: str | None
    help: str | None
    help_behavior_hint: str | Callable[[str], str] | None
    aliases: tuple[str, ...] | None
    prefix_name: bool | None
    constructor_factory: Callable[[], type | Callable[..., Any]] | None
    default: Any = MISSING_NONPROP


@overload
def arg(
    *,
    name: str | None = None,
    metavar: str | None = None,
    help: str | None = None,
    help_behavior_hint: str | Callable[[str], str] | None = None,
    aliases: tuple[str, ...] | list[str] | None = None,
    prefix_name: bool | None = None,
    constructor: None = None,
    constructor_factory: Callable[[], type | Callable[..., Any]] | None = None,
    default: Any = MISSING_NONPROP,
) -> object: ...


@overload
def arg(
    *,
    name: str | None = None,
    metavar: str | None = None,
    help: str | None = None,
    help_behavior_hint: str | Callable[[str], str] | None = None,
    aliases: tuple[str, ...] | list[str] | None = None,
    prefix_name: bool | None = None,
    constructor: type | Callable[..., Any] | None = None,
    constructor_factory: None = None,
    default: Any = MISSING_NONPROP,
) -> object: ...


def arg(
    *,
    name: str | None = None,
    metavar: str | None = None,
    help: str | None = None,
    help_behavior_hint: str | Callable[[str], str] | None = None,
    aliases: tuple[str, ...] | list[str] | None = None,
    prefix_name: bool | None = None,
    constructor: type | Callable[..., Any] | None = None,
    constructor_factory: Callable[[], type | Callable[..., Any]] | None = None,
    default: Any = MISSING_NONPROP,
) -> object:
    """Provides fine-grained control over individual CLI argument properties.

    The `arg()` function allows you to customize how individual arguments appear and
    behave in the command-line interface. This provides more control than relying on
    the automatic argument generation.

    Example::

        from dataclasses import dataclass
        from typing import Annotated
        import tyro

        @dataclass
        class Config:
            # Default argument appearance
            regular_option: int = 1

            # Customized argument
            custom_option: Annotated[
                str,
                tyro.conf.arg(
                    name="opt",                     # Shorter name
                    help="Custom help message",     # Override docstring
                    aliases=("-o", "--short-opt"),  # Alternative flags
                    metavar="VALUE"                 # Display in help
                )
            ] = "default"

        # Usage:
        # python script.py --regular-option 5 --opt custom_value
        # python script.py --regular-option 5 -o custom_value

    The `arg()` function should be used at the root level of annotations and not
    nested within container types like lists.

    Args:
        name: A custom name for the argument in the CLI.
        metavar: Argument placeholder shown in usage messages. The type is used by default.
        help: Custom helptext for this argument. The docstring is used by default.
        help_behavior_hint: Override the highlighted hint text that follows the helptext.
            This is typically used for hints like "(default: XXX)" or "(optional)".
            You can provide either a string or a lambda function that takes a formatted
            default value as input.
        aliases: Alternative flag names for this argument. All strings must start
            with a hyphen (-). Aliases are not prefixed in nested structures and are
            not supported for positional arguments.
        prefix_name: Controls whether to prefix the argument name based on its position
            in a nested structure. Arguments are prefixed by default.
        constructor: A custom constructor type or function to use in place of the
            argument's type for parsing. See :mod:`tyro.constructors` for more details.
        constructor_factory: A function that returns a constructor type for parsing.
            This cannot be used together with the constructor parameter.
        default: Default value for the argument. This will be used only if the field
            does not have a default value. The field default takes precedence.

    Returns:
        A configuration object that should be attached to a type using `Annotated[]`.
    """
    assert not (constructor is not None and constructor_factory is not None), (
        "`constructor` and `constructor_factory` cannot both be set."
    )

    if aliases is not None:
        for alias in aliases:
            assert alias.startswith("-"), "Argument alias needs to start with a hyphen!"

    return _ArgConfig(
        name=name,
        metavar=metavar,
        help=help,
        help_behavior_hint=help_behavior_hint,
        aliases=tuple(aliases) if aliases is not None else None,
        prefix_name=prefix_name,
        constructor_factory=(
            constructor_factory if constructor is None else lambda: constructor
        ),
        default=default,
    )



"""Choices

:py:data:`typing.Literal[]` can be used to restrict inputs to a fixed set of literal choices.

Usage:

    python ./05_choices.py --help
    python ./05_choices.py --string red
    python ./05_choices.py --string blue
"""

import dataclasses
from pprint import pprint
from typing import Literal

import tyro


@dataclasses.dataclass
class Args:
    # We can use Literal[] to restrict the set of allowable inputs, for example, over
    # a set of strings.
    string: Literal["red", "green"] = "red"

    # Integers also work. (as well as booleans, enums, etc)
    number: Literal[0, 1, 2] = 0


if __name__ == "__main__":
    args = tyro.cli(Args)
    pprint(args)



from __future__ import annotations

import dataclasses
import functools
import warnings
from typing import Any, cast

from .. import _docstrings, _resolver
from .._singleton import (
    MISSING,
    MISSING_NONPROP,
    is_missing,
)
from ._struct_spec import StructConstructorSpec, StructFieldSpec, StructTypeInfo
from ._struct_spec_flax import is_flax_module


def _ensure_dataclass_instance_used_as_default_is_frozen(
    field: dataclasses.Field, default_instance: Any
) -> None:
    """Ensure that a dataclass type used directly as a default value is marked as
    frozen."""
    assert dataclasses.is_dataclass(default_instance)
    cls = type(default_instance)
    if not cls.__dataclass_params__.frozen:  # type: ignore
        warnings.warn(
            f"Mutable type {cls} is used as a default value for `{field.name}`. This is"
            " dangerous! Consider using `dataclasses.field(default_factory=...)` or"
            f" marking {cls} as frozen."
        )


def _get_dataclass_field_default(
    field: dataclasses.Field, parent_default_instance: Any
) -> Any:
    """Helper for getting the default instance for a dataclass field."""
    # If the dataclass's parent is explicitly marked MISSING, mark this field as missing
    # as well.
    if parent_default_instance is MISSING:
        return MISSING

    # Try grabbing default from parent instance.
    if not is_missing(parent_default_instance) and parent_default_instance is not None:
        # Populate default from some parent, eg `default=` in `tyro.cli()`.
        if hasattr(parent_default_instance, field.name):
            return getattr(parent_default_instance, field.name)

    # Try grabbing default from dataclass field.
    if field.default is not dataclasses.MISSING:
        default = field.default
        # dataclasses.is_dataclass() will also return true for dataclass
        # _types_, not just instances.
        if type(default) is not type and dataclasses.is_dataclass(default):
            _ensure_dataclass_instance_used_as_default_is_frozen(field, default)
        return default

    # Populate default from `dataclasses.field(default_factory=...)`.
    if field.default_factory is not dataclasses.MISSING and not (
        # Special case to ignore default_factory if we write:
        # `field: Dataclass = dataclasses.field(default_factory=Dataclass)`.
        #
        # In other words, treat it the same way as: `field: Dataclass`.
        #
        # The only time this matters is when we our dataclass has a `__post_init__`
        # function that mutates the dataclass. We choose here to use the default values
        # before this method is called.
        dataclasses.is_dataclass(field.type) and field.default_factory is field.type
    ):
        return field.default_factory()

    # Otherwise, no default.
    return MISSING_NONPROP


def dataclass_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
    """Rule for handling dataclass types."""
    if not dataclasses.is_dataclass(info.type):
        return None

    # Check if this is a flax module and get fields to skip
    is_flax, flax_skip_fields = is_flax_module(info.type)

    # Check if this is a Pydantic dataclass (which has different init=False semantics).
    # Pydantic dataclasses have __pydantic_config__ attribute.
    is_pydantic_dataclass = hasattr(info.type, "__pydantic_config__")

    # Handle dataclasses.
    field_list = []
    init_false_field_names: set[str] = set()
    for dc_field in _resolver.resolved_fields(info.type):
        # For flax modules, we ignore the built-in fields.
        if is_flax and dc_field.name in flax_skip_fields:
            continue

        # Check if this field should be excluded from __init__.
        # For standard dataclasses, this is determined by dc_field.init.
        # For Pydantic dataclasses, we also need to check the FieldInfo.init attribute,
        # because Pydantic dataclasses always report dc_field.init=True even when
        # Field(init=False) is used.
        field_should_init = dc_field.init
        if is_pydantic_dataclass and field_should_init:
            # For Pydantic dataclasses, check if the field has a FieldInfo with init=False.
            # The FieldInfo object is stored in dc_field.default.
            if getattr(dc_field.default, "init", None) is False:
                field_should_init = False

        # Handle init=False fields specially.
        if not field_should_init:
            # For init=False fields, we can't pass them to the constructor.
            # Only include them if a default instance is provided with a value.
            if not is_missing(info.default) and hasattr(info.default, dc_field.name):
                # Use value from default instance.
                init_false_field_names.add(dc_field.name)
                default = getattr(info.default, dc_field.name)
            else:
                # No default instance value, skip this field entirely.
                continue
        else:
            default = _get_dataclass_field_default(dc_field, info.default)

        # Try to get helptext from field metadata. This is also intended to be
        # compatible with HuggingFace-style config objects.
        helptext = dc_field.metadata.get("help", None)
        assert isinstance(helptext, (str, type(None)))

        # Try to get helptext from docstrings. This can't be generated
        # dynamically.
        if helptext is None:
            # Lazy docstring parsing: use partial to defer expensive parsing.
            helptext = functools.partial(
                _docstrings.get_field_docstring,
                info.type,
                dc_field.name,
                info.markers,
            )

        assert not isinstance(dc_field.type, str)

        field_list.append(
            StructFieldSpec(
                name=dc_field.name,
                type=cast(Any, dc_field.type),
                default=default,
                helptext=helptext,
            )
        )

    # Wrap the instantiate function if we have init=False fields to exclude from call.
    instantiate = info.type
    if len(init_false_field_names) > 0:

        def wrapped_instantiate(**kwargs):
            # Remove init=False fields from kwargs and save their values.
            init_false_values = {
                k: kwargs.pop(k) for k in init_false_field_names if k in kwargs
            }

            # Call the constructor without init=False fields.
            instance = info.type(**kwargs)

            # Set the init=False field values on the instance.
            # Use object.__setattr__ to bypass frozen dataclass protection.
            for field_name, value in init_false_values.items():
                object.__setattr__(instance, field_name, value)

            return instance

        instantiate = wrapped_instantiate

    return StructConstructorSpec(
        instantiate=instantiate,
        fields=tuple(field_list),
    )



"""Tests for experimental features."""

import dataclasses
import sys
from io import StringIO

import tyro


def test_enable_timing() -> None:
    """Test that timing can be enabled and produces output."""

    @dataclasses.dataclass
    class Args:
        x: int = 5

    # Enable timing.
    tyro._experimental_options["enable_timing"] = True

    # Capture stderr to check for timing output.
    old_stderr = sys.stderr
    sys.stderr = StringIO()

    try:
        result = tyro.cli(Args, args=["--x", "10"])
        assert result.x == 10

        # Check that timing output was produced.
        stderr_output = sys.stderr.getvalue()
        assert "took" in stderr_output
        assert "seconds" in stderr_output
    finally:
        # Restore stderr and disable timing.
        sys.stderr = old_stderr
        tyro._experimental_options["enable_timing"] = False



"""Backtracking parser for handling variable-length argument sequences."""

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ._primitive_spec import PrimitiveConstructorSpec


@dataclasses.dataclass
class BacktrackState:
    """State for backtracking parser."""

    spec_idx: int
    arg_idx: int
    parsed_value: Any | None  # The value parsed at this state, if any.
    parent: BacktrackState | None  # Link to parent state to reconstruct path.
    nargs_option_idx: int


def parse_with_backtracking(
    args: list[str],
    specs: tuple[PrimitiveConstructorSpec[Any], ...],
    is_repeating: bool = False,
) -> list[Any] | None:
    """Parse arguments using backtracking when specs have variable nargs.

    Args:
        args: List of string arguments to parse.
        specs: Tuple of PrimitiveConstructorSpec instances.
        is_repeating: If True with single spec, use specs[0] repeatedly.
            If True with multiple specs, repeat the entire sequence of specs.
            If False, use each spec in order exactly once.

    Returns:
        List of parsed values if successful, None if no valid parse exists.

    Examples:
        # For sequences like List[Union[int, Tuple[int, int]]]:
        result = parse_with_backtracking(
            args=["1", "2", "3", "4"],
            specs=(spec_with_nargs_1_or_2,),
            is_repeating=True
        )

        # For dicts like Dict[str, Union[int, Tuple[int, int]]]:
        result = parse_with_backtracking(
            args=["key1", "1", "2", "key2", "3"],
            specs=(key_spec, val_spec),
            is_repeating=True
        )

        # For tuples like Tuple[Union[int, str], Union[float, Tuple[float, float]]]:
        result = parse_with_backtracking(
            args=["hello", "1.5", "2.5"],
            specs=(int_or_str_spec, float_or_pair_spec),
            is_repeating=False
        )
    """

    assert len(specs) >= 1, "At least one spec is required"

    # Use iterative approach with explicit stack to avoid recursion limit.
    stack: list[BacktrackState] = [BacktrackState(0, 0, None, None, 0)]

    def reconstruct_path(state: BacktrackState) -> list[Any]:
        """Reconstruct the parsed values from the state chain."""
        result = []
        current: BacktrackState | None = state
        while current is not None:
            if current.parent is not None:  # Skip the initial state
                result.append(current.parsed_value)
            current = current.parent
        return list(reversed(result))

    while stack:
        state = stack.pop()
        spec_idx = state.spec_idx
        arg_idx = state.arg_idx
        nargs_option_idx = state.nargs_option_idx

        # Base cases.
        if is_repeating:
            # For repeating specs, success when all args consumed.
            if arg_idx == len(args):
                # When repeating multiple specs, ensure we completed full cycles.
                if len(specs) > 1 and spec_idx % len(specs) != 0:
                    continue  # Incomplete cycle, not a valid parse.
                return reconstruct_path(state)
            spec = specs[spec_idx % len(specs)]
        else:
            # For non-repeating specs, success when all specs processed.
            if spec_idx == len(specs):
                if arg_idx == len(args):
                    return reconstruct_path(state)
                else:
                    continue
            if arg_idx == len(args):
                continue
            spec = specs[spec_idx]

        # Get nargs options for current spec.
        if spec.nargs == "*":
            # For nargs='*', try all possible lengths from 0 to remaining args.
            # Try longer matches first to prefer consuming more args.
            nargs_options = tuple(range(len(args) - arg_idx, -1, -1))
        else:
            nargs_options = (spec.nargs,) if isinstance(spec.nargs, int) else spec.nargs

        # Check if we've tried all nargs options for this state.
        if nargs_option_idx >= len(nargs_options):
            continue

        # Push state for trying next nargs option.
        if nargs_option_idx + 1 < len(nargs_options):
            stack.append(
                BacktrackState(
                    spec_idx,
                    arg_idx,
                    state.parsed_value,
                    state.parent,
                    nargs_option_idx + 1,
                )
            )

        # Try current nargs option.
        nargs_option = nargs_options[nargs_option_idx]
        if arg_idx + nargs_option > len(args):
            continue

        # Extract candidate arguments.
        candidate_args = args[arg_idx : arg_idx + nargs_option]
        if spec.choices is not None and any(
            arg not in spec.choices for arg in candidate_args
        ):
            continue

        try:
            # Try to parse this chunk.
            parsed = spec.instance_from_str(candidate_args)
            next_spec_idx = spec_idx + 1

            # Push new state to explore this path.
            stack.append(
                BacktrackState(
                    next_spec_idx,
                    arg_idx + nargs_option,
                    parsed,
                    state,
                    0,
                )
            )
        except ValueError:
            # This option didn't work, will try next via the pushed state above.
            continue

    # No valid parse found.
    return None



"""Tuples and ``NamedTuple``

Example using :func:`tyro.cli()` to instantiate tuple types. :py:class:`tuple`,
:py:data:`typing.Tuple`, and :py:class:`typing.NamedTuple` are all supported.

Usage:

    python ./05_tuples.py --help
    python ./05_tuples.py --color 127 127 127
    python ./05_tuples.py --two-colors.1.r 127 --two-colors.1.g 0 --two-colors.1.b 0
"""

from typing import NamedTuple

import tyro


# Named tuples are interpreted as nested structures.
class Color(NamedTuple):
    r: int
    g: int
    b: int


class TupleType(NamedTuple):
    """Description.
    This should show up in the helptext!"""

    # Tuple types can contain raw values.
    color: tuple[int, int, int] = (255, 0, 0)

    # Tuple types can contain nested structures.
    two_colors: tuple[Color, Color] = (Color(255, 0, 0), Color(0, 255, 0))


if __name__ == "__main__":
    x = tyro.cli(TupleType)
    assert isinstance(x, tuple)
    print(x)



"""Tests for Python eval-based collections.

These tests verify that tyro can parse collections using Python syntax
(e.g., "[128, 128, 128]", "{'a': 1}") via eval() when the
UsePythonSyntaxForLiteralCollections marker is used.

The eval() approach supports both literal values and non-literal types like
pathlib.Path by making common types available in a secure eval() context.

This is particularly useful for wandb sweeps integration.
"""

from dataclasses import dataclass
from typing import Dict, List, Literal, Optional, Set, Tuple

import pytest

import tyro


def test_tuple_python_syntax() -> None:
    """Test that Python literal syntax works for tuples."""

    @dataclass
    class Config:
        dims: Tuple[int, int, int]

    # Normal usage without marker.
    assert tyro.cli(Config, args=["--dims", "1", "2", "3"]) == Config(dims=(1, 2, 3))

    # With UsePythonSyntaxForLiteralCollections marker.
    assert tyro.cli(
        Config,
        args=["--dims", "(1, 2, 3)"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    ) == Config(dims=(1, 2, 3))


def test_tuple_strings() -> None:
    """Test Python syntax for tuples with string elements."""

    @dataclass
    class Config:
        names: Tuple[str, str]

    assert tyro.cli(
        Config,
        args=["--names", "('foo', 'bar')"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    ) == Config(names=("foo", "bar"))


def test_list_python_syntax() -> None:
    """Test Python syntax for lists."""

    @dataclass
    class Config:
        values: List[int]

    # Normal usage without marker.
    assert tyro.cli(Config, args=["--values", "1", "2", "3"]) == Config(
        values=[1, 2, 3]
    )

    # With UsePythonSyntaxForLiteralCollections marker.
    assert tyro.cli(
        Config,
        args=["--values", "[1, 2, 3]"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    ) == Config(values=[1, 2, 3])


def test_dict_python_syntax() -> None:
    """Test Python syntax for dicts."""

    @dataclass
    class Config:
        mapping: Dict[str, int]

    # Normal usage without marker.
    result = tyro.cli(Config, args=["--mapping", "a", "1", "b", "2"])
    assert result == Config(mapping={"a": 1, "b": 2})

    # With UsePythonSyntaxForLiteralCollections marker.
    result = tyro.cli(
        Config,
        args=["--mapping", "{'a': 1, 'b': 2}"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result == Config(mapping={"a": 1, "b": 2})


def test_set_python_syntax() -> None:
    """Test Python syntax for sets."""

    @dataclass
    class Config:
        values: Set[int]

    result = tyro.cli(
        Config,
        args=["--values", "{1, 2, 3}"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result == Config(values={1, 2, 3})


def test_nested_structure() -> None:
    """Test nested structures with Python syntax."""

    @dataclass
    class Inner:
        size: Tuple[int, int]

    @dataclass
    class Outer:
        inner: Inner

    result = tyro.cli(
        Outer,
        args=["--inner.size", "(100, 200)"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result == Outer(inner=Inner(size=(100, 200)))


def test_mixed_types() -> None:
    """Test tuple with mixed numeric types."""

    @dataclass
    class Config:
        mixed: Tuple[int, float, int]

    result = tyro.cli(
        Config,
        args=["--mixed", "(1, 2.5, 3)"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result == Config(mixed=(1, 2.5, 3))


def test_invalid_python_syntax() -> None:
    """Test that invalid Python syntax raises an error."""

    @dataclass
    class Config:
        values: List[int]

    with pytest.raises(SystemExit):
        tyro.cli(
            Config,
            args=["--values", "[1, 2, "],
            config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
        )


def test_type_mismatch() -> None:
    """Test that type mismatches raise an error."""

    @dataclass
    class Config:
        values: Tuple[int, int]

    # Passing a list when expecting a tuple.
    with pytest.raises(SystemExit):
        tyro.cli(
            Config,
            args=["--values", "[1, 2]"],
            config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
        )


def test_helptext_coverage() -> None:
    """Test helptext generation to cover str_from_instance."""
    import sys
    from dataclasses import field

    @dataclass
    class Config:
        simple: List[int] = field(default_factory=lambda: [1, 2, 3])
        nested: List[Tuple[int, str]] = field(
            default_factory=lambda: [(1, "a"), (2, "b")]
        )

    # Getting helptext will call str_from_instance for defaults.
    sys.argv = ["test", "--help"]
    with pytest.raises(SystemExit):
        tyro.cli(Config, config=(tyro.conf.UsePythonSyntaxForLiteralCollections,))


def test_union_with_default() -> None:
    """Test union types with defaults to cover is_instance_fn."""
    import sys

    @dataclass
    class Config:
        value: Tuple[int, ...] | int = 3

    # Getting helptext with a union default exercises is_instance.
    sys.argv = ["test", "--help"]
    with pytest.raises(SystemExit):
        tyro.cli(Config, config=(tyro.conf.UsePythonSyntaxForLiteralCollections,))


def test_variable_length_tuple() -> None:
    """Test variable-length tuples."""

    @dataclass
    class Config:
        values: Tuple[int, ...]

    result = tyro.cli(
        Config,
        args=["--values", "(1, 2, 3, 4, 5)"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result == Config(values=(1, 2, 3, 4, 5))


def test_nested_union_metavar() -> None:
    """Test metavar generation for nested union types to cover edge cases."""
    import sys

    @dataclass
    class Config:
        # This will exercise the metavar generation with Union types.
        values: List[int | str]

    # Getting helptext will generate metavar for the type.
    sys.argv = ["test", "--help"]
    with pytest.raises(SystemExit):
        tyro.cli(Config, config=(tyro.conf.UsePythonSyntaxForLiteralCollections,))


def test_nested_literal_metavar() -> None:
    """Test metavar generation for Literal types (which lack __name__)."""
    import sys

    @dataclass
    class Config:
        # Literal types don't have __name__, testing edge case in metavar generation.
        values: List[Literal["a", "b"]]

    # Getting helptext will generate metavar for the type.
    sys.argv = ["test", "--help"]
    with pytest.raises(SystemExit):
        tyro.cli(Config, config=(tyro.conf.UsePythonSyntaxForLiteralCollections,))


def test_pathlib_fallback() -> None:
    """Test that Path types fall back to normal handling (not literal-eval compatible)."""
    from pathlib import Path

    @dataclass
    class Config:
        paths: List[Path]

    # Path is not compatible with ast.literal_eval, so this should use normal
    # multi-arg parsing even with the marker present.
    result = tyro.cli(
        Config,
        args=["--paths", "foo", "bar"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result == Config(paths=[Path("foo"), Path("bar")])


def test_unparameterized_generic_fallback() -> None:
    """Test that unparameterized generics fall back to other rules."""
    # When a collection type has no type args, the eval rule should return None
    # and let other rules handle it. typing.List without subscript has origin=list
    # but no args, which triggers this early return path.
    from typing import List

    @dataclass
    class Config:
        # Using unparameterized typing.List triggers the early return.
        # It will be handled by other rules as a list with inferred type.
        values: List = None  # type: ignore

    # This should work - the eval rule returns None, other rules handle it.
    result = tyro.cli(
        Config,
        args=[],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.values is None


def test_incompatible_type_fallback() -> None:
    """Test that incompatible types are properly validated and marker is ignored."""
    # When collection contains non-eval-compatible types, the marker is ignored.
    from dataclasses import dataclass as inner_dataclass
    from dataclasses import field

    @inner_dataclass
    class Inner:
        value: int

    @dataclass
    class Config:
        # Inner is a struct type (dataclass), so marker is ignored.
        # Tyro falls back to normal handling (requires explicit default).
        items: List[Inner] = field(default_factory=list)

    # With empty list default, tyro should use it.
    result = tyro.cli(
        Config,
        args=[],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.items == []


def test_literal_type_support() -> None:
    """Test that Literal types work with the marker."""

    @dataclass
    class Config:
        values: List[Literal["a", "b", "c"]]

    # With marker, should accept Python literal syntax.
    result = tyro.cli(
        Config,
        args=["--values", "['a', 'b', 'c']"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.values == ["a", "b", "c"]


def test_literal_int_support() -> None:
    """Test that Literal with int values works."""

    @dataclass
    class Config:
        values: Tuple[Literal[1, 2, 3], Literal[1, 2, 3]]

    result = tyro.cli(
        Config,
        args=["--values", "(1, 2)"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.values == (1, 2)


def test_optional_with_literal() -> None:
    """Test Optional[Literal[...]] works."""

    @dataclass
    class Config:
        value: Optional[List[Literal["x", "y", "z"]]] = None

    result = tyro.cli(
        Config,
        args=[],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.value is None

    result = tyro.cli(
        Config,
        args=["--value", "['x', 'y']"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.value == ["x", "y"]


def test_custom_class_fallback() -> None:
    """Test that custom classes fall back to normal handling."""
    from dataclasses import field

    class CustomClass:
        def __init__(self, value: int):
            self.value = value

    @dataclass
    class Config:
        # CustomClass is not a built-in, so marker should be ignored.
        # Tyro falls back to normal handling (treating as struct with default).
        items: List[CustomClass] = field(default_factory=list)

    # With an empty list default, tyro should use it.
    result = tyro.cli(
        Config,
        args=[],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.items == []


def test_unsupported_builtins_fallback() -> None:
    """Test that unsupported built-in types (frozenset, range, slice) fall back."""
    from dataclasses import field

    @dataclass
    class Config:
        # frozenset is a built-in type but not supported by ast.literal_eval().
        # Tyro falls back to normal handling (treating as struct with default).
        values: List[frozenset] = field(default_factory=list)

    # With an empty list default, tyro should use it.
    result = tyro.cli(
        Config,
        args=[],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.values == []


def test_complex_number_support() -> None:
    """Test that complex numbers work with the marker."""

    @dataclass
    class Config:
        values: List[complex]

    result = tyro.cli(
        Config,
        args=["--values", "[1+2j, 3+4j, 5+0j]"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.values == [1 + 2j, 3 + 4j, 5 + 0j]


def test_literal_metavar_formatting() -> None:
    """Test that Literal types are formatted correctly in metavar."""
    import sys

    @dataclass
    class Config:
        # String literals.
        str_literals: List[Literal["a", "b", "c"]]
        # Int literals.
        int_literals: Tuple[Literal[1, 2, 3], Literal[4, 5, 6]]
        # Nested with Literal.
        nested: List[Tuple[Literal["x", "y"], int]]

    # Getting helptext will show the metavar.
    sys.argv = ["test", "--help"]
    with pytest.raises(SystemExit):
        tyro.cli(Config, config=(tyro.conf.UsePythonSyntaxForLiteralCollections,))


def test_bytes_support() -> None:
    """Test that bytes work with the marker."""

    @dataclass
    class Config:
        values: List[bytes]

    result = tyro.cli(
        Config,
        args=["--values", "[b'hello', b'world']"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.values == [b"hello", b"world"]


def test_mixed_literal_and_regular_types() -> None:
    """Test mixing Literal types with regular types in collections."""

    @dataclass
    class Config:
        # Tuple with Literal and regular types.
        mixed: Tuple[Literal["a", "b"], int, str]

    result = tyro.cli(
        Config,
        args=["--mixed", "('a', 42, 'hello')"],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.mixed == ("a", 42, "hello")


def test_none_in_tuple() -> None:
    """Test that None works correctly in tuples."""

    @dataclass
    class Config:
        # Tuple with None.
        value: Tuple[int, None]
        # List of tuples with None.
        values: List[Tuple[str, None]]

    result = tyro.cli(
        Config,
        args=[
            "--value",
            "(42, None)",
            "--values",
            "[('a', None), ('b', None)]",
        ],
        config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
    )
    assert result.value == (42, None)
    assert result.values == [("a", None), ("b", None)]


def test_none_metavar() -> None:
    """Test that None is formatted as 'None' in metavar (not 'NONETYPE')."""
    import sys

    @dataclass
    class Config:
        value: Tuple[int, None]

    # Getting helptext will show the metavar.
    sys.argv = ["test", "--help"]
    with pytest.raises(SystemExit):
        tyro.cli(Config, config=(tyro.conf.UsePythonSyntaxForLiteralCollections,))



# mypy: ignore-errors
"""Tests for ml_collections integration."""

import dataclasses

from helptext_utils import get_helptext_with_checks

import tyro


def test_basic_configdict() -> None:
    """Test basic ConfigDict parsing."""
    from ml_collections import ConfigDict

    def get_config() -> ConfigDict:
        config = ConfigDict()
        config.wandb = ConfigDict()
        config.wandb.mode = "online"
        config.wandb.project = "test-project"

        config.network = ConfigDict()
        config.network.policy_layer_dims = (128,) * 3
        config.network.value_layer_dims = (256,) * 5
        return config

    def train(config: ConfigDict = get_config()) -> ConfigDict:
        return config  # type: ignore

    # Test parsing with CLI arguments
    result = tyro.cli(
        train,
        args=[
            "--config.wandb.mode=offline",
            "--config.network.policy-layer-dims=64",
        ],
    )
    assert result.wandb.mode == "offline"  # type: ignore
    assert result.wandb.project == "test-project"  # type: ignore
    assert result.network.policy_layer_dims == (64,)  # type: ignore
    assert result.network.value_layer_dims == (256, 256, 256, 256, 256)  # type: ignore

    # Test helptext
    helptext = get_helptext_with_checks(train)
    assert "--config.wandb.mode" in helptext
    assert "--config.wandb.project" in helptext
    assert "--config.network.policy-layer-dims" in helptext
    assert "--config.network.value-layer-dims" in helptext


def test_field_references() -> None:
    """Test ConfigDict with FieldReference."""
    from ml_collections import ConfigDict, FieldReference

    def get_config() -> ConfigDict:
        config = ConfigDict()
        layer_dim_ref = FieldReference((128,) * 3)
        config.layer_dims = layer_dim_ref

        config.network = ConfigDict()
        config.network.policy_layer_dims = layer_dim_ref
        config.network.value_layer_dims = layer_dim_ref
        return config

    def train(config: ConfigDict = get_config()) -> None:
        return config  # type: ignore

    # Test updating field reference via CLI
    result = tyro.cli(
        train,
        args=[
            "--config.layer-dims=64",
        ],
    )
    assert result.layer_dims == (64,)  # type: ignore
    # Field references also update, so these values should be the same as layer_dims
    assert result.network.policy_layer_dims == (64,)  # type: ignore
    assert result.network.value_layer_dims == (64,)  # type: ignore

    # Test updating directly instead of through reference
    # In this case, updating policy_layer_dims directly updates the reference too
    result = tyro.cli(
        train,
        args=[
            "--config.network.policy-layer-dims=32",
        ],
    )
    # When we update one of the references, the original gets updated too
    assert result.layer_dims == (32,)  # type: ignore
    assert result.network.policy_layer_dims == (32,)  # type: ignore
    assert result.network.value_layer_dims == (32,)  # type: ignore

    # Test helptext
    helptext = get_helptext_with_checks(train)
    assert "--config.layer-dims" in helptext
    assert "Reference default: (128, 128, 128)" in helptext
    assert "(assigns reference)" in helptext


def test_nested_configdict() -> None:
    """Test deeply nested ConfigDict objects."""
    from ml_collections import ConfigDict

    def get_config() -> ConfigDict:
        config = ConfigDict()
        config.level1 = ConfigDict()
        config.level1.level2 = ConfigDict()
        config.level1.level2.level3 = ConfigDict()
        config.level1.level2.level3.value = 42
        config.level1.level2.level3.sequence = ["a", "b", "c"]  # type: ignore
        return config

    def nested_fn(config: ConfigDict = get_config()) -> ConfigDict:
        return config

    # Test updating deeply nested values
    result = tyro.cli(
        nested_fn,
        args=[
            "--config.level1.level2.level3.value=100",
            "--config.level1.level2.level3.sequence=x",
        ],
    )
    assert result.level1.level2.level3.value == 100  # type: ignore

    # Since sequence is a method on ConfigDict, we need to access it differently
    sequence_value = result.level1.level2.level3["sequence"]  # type: ignore
    assert isinstance(sequence_value, list)
    assert len(sequence_value) == 1
    assert sequence_value[0] == "x"

    # Test helptext
    helptext = get_helptext_with_checks(nested_fn)
    assert "--config.level1.level2.level3.value" in helptext
    assert "--config.level1.level2.level3.sequence" in helptext


def test_configdict_in_dataclass() -> None:
    """Test ConfigDict as a field within a dataclass."""
    from ml_collections import ConfigDict

    def get_config() -> ConfigDict:
        config = ConfigDict()
        config.policy_layer_dims = (128,) * 3
        config.value_layer_dims = (256,) * 2
        return config

    @dataclasses.dataclass
    class TrainingConfig:
        learning_rate: float = 0.001
        batch_size: int = 64
        network: ConfigDict = dataclasses.field(default_factory=get_config)

    # Create a test function that uses the dataclass with a configdict
    def train_with_config(config: ConfigDict = get_config()) -> ConfigDict:
        return config

    # Test with pure ConfigDict
    result = tyro.cli(
        train_with_config,
        args=[
            "--config.policy-layer-dims=64",
        ],
    )
    assert result.policy_layer_dims == (64,)
    assert result.value_layer_dims == (256, 256)

    # Test helptext
    helptext = get_helptext_with_checks(train_with_config)
    assert "--config.policy-layer-dims" in helptext
    assert "--config.value-layer-dims" in helptext


def test_configdict_with_tuples() -> None:
    """Test ConfigDict with tuple fields of different types."""
    from ml_collections import ConfigDict

    def get_config() -> ConfigDict:
        config = ConfigDict()
        config.int_tuple = (1, 2, 3)
        config.float_tuple = (1.0, 2.0, 3.0)
        config.str_tuple = ("a", "b", "c")
        config.mixed_tuple = (1, "a", 2.0)
        return config

    def process(config: ConfigDict = get_config()) -> ConfigDict:
        return config

    # Test various tuple types
    result = tyro.cli(
        process,
        args=[
            "--config.int-tuple=10",
            "--config.float-tuple=1.5",
            "--config.str-tuple=x",
            "--config.mixed-tuple",
            "5",
            "3",
            "1",
        ],
    )
    assert result.int_tuple == (10,)
    assert result.float_tuple == (1.5,)
    assert result.str_tuple == ("x",)
    assert result.mixed_tuple == (5, "3", 1.0)

    # Test helptext
    helptext = get_helptext_with_checks(process)
    assert "--config.int-tuple" in helptext
    assert "--config.float-tuple" in helptext
    assert "--config.str-tuple" in helptext


# FrozenConfigDict Tests


def test_basic_frozenconfigdict() -> None:
    """Test basic FrozenConfigDict parsing."""
    from ml_collections import FrozenConfigDict

    def get_config() -> FrozenConfigDict:
        config_dict = {}
        config_dict["wandb"] = {}
        config_dict["wandb"]["mode"] = "online"
        config_dict["wandb"]["project"] = "test-project"

        config_dict["network"] = {}
        config_dict["network"]["policy_layer_dims"] = (128,) * 3
        config_dict["network"]["value_layer_dims"] = (256,) * 5
        return FrozenConfigDict(config_dict)

    def train(config: FrozenConfigDict = get_config()) -> FrozenConfigDict:
        return config  # type: ignore

    # Test parsing with CLI arguments
    result = tyro.cli(
        train,
        args=[
            "--config.wandb.mode=offline",
            "--config.network.policy-layer-dims=64",
        ],
    )
    assert result.wandb.mode == "offline"  # type: ignore
    assert result.wandb.project == "test-project"  # type: ignore
    assert result.network.policy_layer_dims == (64,)  # type: ignore
    assert result.network.value_layer_dims == (256, 256, 256, 256, 256)  # type: ignore

    # Test helptext
    helptext = get_helptext_with_checks(train)
    assert "--config.wandb.mode" in helptext
    assert "--config.wandb.project" in helptext
    assert "--config.network.policy-layer-dims" in helptext
    assert "--config.network.value-layer-dims" in helptext


def test_field_references_frozenconfigdict() -> None:
    """Test FrozenConfigDict with FieldReference."""
    from ml_collections import FieldReference, FrozenConfigDict

    def get_config() -> FrozenConfigDict:
        config_dict = {}
        layer_dim_ref = FieldReference((128,) * 3)
        config_dict["layer_dims"] = layer_dim_ref

        config_dict["network"] = {}
        config_dict["network"]["policy_layer_dims"] = layer_dim_ref
        config_dict["network"]["value_layer_dims"] = layer_dim_ref
        return FrozenConfigDict(config_dict)

    def train(config: FrozenConfigDict = get_config()) -> None:
        return config  # type: ignore

    # Test updating field reference via CLI
    result = tyro.cli(
        train,
        args=[
            "--config.layer-dims=64",
        ],
    )
    assert result.layer_dims == (64,)  # type: ignore
    # Field references also update, so these values should be the same as layer_dims
    assert result.network.policy_layer_dims == (64,)  # type: ignore
    assert result.network.value_layer_dims == (64,)  # type: ignore

    # Test updating directly instead of through reference
    # In this case, updating policy_layer_dims directly updates the reference too
    result = tyro.cli(
        train,
        args=[
            "--config.network.policy-layer-dims=32",
        ],
    )
    # When we update one of the references, the original gets updated too
    assert result.layer_dims == (32,)  # type: ignore
    assert result.network.policy_layer_dims == (32,)  # type: ignore
    assert result.network.value_layer_dims == (32,)  # type: ignore

    # Test helptext
    helptext = get_helptext_with_checks(train)
    assert "--config.layer-dims" in helptext
    assert "Reference default: (128, 128, 128)" in helptext
    assert "(assigns reference)" in helptext


def test_nested_frozenconfigdict() -> None:
    """Test deeply nested FrozenConfigDict objects."""
    from ml_collections import FrozenConfigDict

    def get_config() -> FrozenConfigDict:
        config_dict = {}
        config_dict["level1"] = {}
        config_dict["level1"]["level2"] = {}
        config_dict["level1"]["level2"]["level3"] = {}
        config_dict["level1"]["level2"]["level3"]["value"] = 42
        config_dict["level1"]["level2"]["level3"]["sequence"] = ["a", "b", "c"]
        return FrozenConfigDict(config_dict)

    def nested_fn(config: FrozenConfigDict = get_config()) -> FrozenConfigDict:
        return config

    # Test updating deeply nested values
    result = tyro.cli(
        nested_fn,
        args=[
            "--config.level1.level2.level3.value=100",
            "--config.level1.level2.level3.sequence=x",
        ],
    )
    assert result.level1.level2.level3.value == 100  # type: ignore

    # Access sequence
    sequence_value = result.level1.level2.level3.sequence  # type: ignore
    assert isinstance(sequence_value, tuple)  # Will be cast to tuple because frozen!
    assert len(sequence_value) == 1
    assert sequence_value[0] == "x"

    # Test helptext
    helptext = get_helptext_with_checks(nested_fn)
    assert "--config.level1.level2.level3.value" in helptext
    assert "--config.level1.level2.level3.sequence" in helptext


def test_frozenconfigdict_in_dataclass() -> None:
    """Test FrozenConfigDict as a field within a dataclass."""
    from ml_collections import FrozenConfigDict

    def get_config() -> FrozenConfigDict:
        config_dict = {}
        config_dict["policy_layer_dims"] = (128,) * 3
        config_dict["value_layer_dims"] = (256,) * 2
        return FrozenConfigDict(config_dict)

    @dataclasses.dataclass
    class TrainingConfig:
        learning_rate: float = 0.001
        batch_size: int = 64
        network: FrozenConfigDict = dataclasses.field(default_factory=get_config)

    # Create a test function that uses the dataclass with a FrozenConfigDict
    def train_with_config(config: FrozenConfigDict = get_config()) -> FrozenConfigDict:
        return config

    # Test with pure FrozenConfigDict
    result = tyro.cli(
        train_with_config,
        args=[
            "--config.policy-layer-dims=64",
        ],
    )
    assert result.policy_layer_dims == (64,)
    assert result.value_layer_dims == (256, 256)

    # Test helptext
    helptext = get_helptext_with_checks(train_with_config)
    assert "--config.policy-layer-dims" in helptext
    assert "--config.value-layer-dims" in helptext


def test_frozenconfigdict_with_tuples() -> None:
    """Test FrozenConfigDict with tuple fields of different types."""
    from ml_collections import FrozenConfigDict

    def get_config() -> FrozenConfigDict:
        config_dict = {}
        config_dict["int_tuple"] = (1, 2, 3)
        config_dict["float_tuple"] = (1.0, 2.0, 3.0)
        config_dict["str_tuple"] = ("a", "b", "c")
        config_dict["mixed_tuple"] = (1, "a", 2.0)
        return FrozenConfigDict(config_dict)

    def process(config: FrozenConfigDict = get_config()) -> FrozenConfigDict:
        return config

    # Test various tuple types
    result = tyro.cli(
        process,
        args=[
            "--config.int-tuple=10",
            "--config.float-tuple=1.5",
            "--config.str-tuple=x",
            "--config.mixed-tuple",
            "5",
            "3",
            "1",
        ],
    )
    assert result.int_tuple == (10,)
    assert result.float_tuple == (1.5,)
    assert result.str_tuple == ("x",)
    assert result.mixed_tuple == (5, "3", 1.0)

    # Test helptext
    helptext = get_helptext_with_checks(process)
    assert "--config.int-tuple" in helptext
    assert "--config.float-tuple" in helptext
    assert "--config.str-tuple" in helptext



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Callable, ClassVar, Union

from tyro._singleton import is_sentinel

from .. import _fmtlib as fmt
from .. import _resolver
from ._primitive_spec import (
    PrimitiveConstructorSpec,
    PrimitiveTypeInfo,
    UnsupportedTypeAnnotationError,
    apply_default_primitive_rules,
)
from ._struct_spec import (
    InvalidDefaultInstanceError,
    StructConstructorSpec,
    StructTypeInfo,
    apply_default_struct_rules,
)

current_registry: ConstructorRegistry | None = None

PrimitiveSpecRule = Callable[
    [PrimitiveTypeInfo],
    Union[PrimitiveConstructorSpec, UnsupportedTypeAnnotationError, None],
]
StructSpecRule = Callable[[StructTypeInfo], Union[StructConstructorSpec, None]]

_check_default_instances_flag: bool = False


def check_default_instances() -> bool:
    """Check whether we should be strict about checking that default types and
    instances match.

    This is usually `False`; tyro attempts to be somewhat lenient when
    inconsistent types are encounted. Strictness, however, is useful for
    matching annotated subcommands to default values.
    """
    return _check_default_instances_flag


@contextmanager
def check_default_instances_context():
    """Context for whether we should be strict about checking that default
    types and instances match.

    This is usually `False`; tyro attempts to be somewhat lenient when
    inconsistent types are encounted. Strictness, however, is useful for
    matching annotated subcommands to default values.
    """
    global _check_default_instances_flag
    old_value = _check_default_instances_flag
    _check_default_instances_flag = True
    try:
        yield
    finally:
        _check_default_instances_flag = old_value


class ConstructorRegistry:
    """Registry for rules that define how types are constructed from
    command-line arguments.

    The behavior of CLIs generated by tyro are based on two types of rules.

    *Primitive rules* should be a callable with the signature:

    .. code-block: python

        (type_info: PrimitiveTypeInfo) -> PrimitiveConstructorSpec | UnsupportedTypeAnnotationError | None

    where `None` is returned if the rule doesn't apply, and
    `UnsupportedTypeAnnotationError` is returned if the rule applies but an
    error was encountered. Each primitive rule defines behavior for a type that
    can be instantiated from a single command-line argument.


    *Struct rules* should be a callable with the signature:

    .. code-block: python

        (type_info: StructTypeInfo) -> StructConstructorSpec | None

    where `None` is returned if the rule doesn't apply. Each struct rule
    defines behavior for a type that can be instantiated from multiple
    command-line arguments.


    To activate a registry, pass it directly to :func:`tyro.cli`:

    .. code-block: python

        registry = ConstructorRegistry()
        tyro.cli(..., registry=registry)

    For backward compatibility, you can also use the context manager pattern, though
    the direct parameter approach is recommended:

    .. code-block: python

        registry = ConstructorRegistry()
        with registry:
            tyro.cli(...)

    """

    _active_registries: ClassVar[list[ConstructorRegistry]] = []

    def __init__(self) -> None:
        self._primitive_rules: list[PrimitiveSpecRule] = []
        self._struct_rules: list[StructSpecRule] = []

    def primitive_rule(self, rule: PrimitiveSpecRule) -> PrimitiveSpecRule:
        """Define a rule for constructing a primitive type from a string. The
        most recently added rule will be applied first.

        Custom primitive rules will take precedence over both default primitive
        rules and struct rules
        """

        self._primitive_rules.append(rule)
        return rule

    def struct_rule(self, rule: StructSpecRule) -> StructSpecRule:
        """Define a rule for constructing a primitive type from a string. The
        most recently added rule will be applied first."""

        self._struct_rules.append(rule)
        return rule

    @classmethod
    def _is_primitive_type(
        cls, type: Any, markers: set[Any], nondefault_only: bool = False
    ) -> bool:
        """Check if a type is a primitive type."""
        return isinstance(
            cls.get_primitive_spec(
                PrimitiveTypeInfo.make(type, markers), nondefault_only=nondefault_only
            ),
            PrimitiveConstructorSpec,
        )

    @classmethod
    def get_primitive_spec(
        cls, type_info: PrimitiveTypeInfo, nondefault_only: bool = False
    ) -> PrimitiveConstructorSpec | UnsupportedTypeAnnotationError:
        """Get a constructor specification for a given type."""

        cls._ensure_defaults_initialized()

        if type_info._primitive_spec is not None:
            return type_info._primitive_spec

        for registry in (
            cls._active_registries[1:] if nondefault_only else cls._active_registries
        )[::-1]:
            for spec_factory in registry._primitive_rules[::-1]:
                maybe_spec = spec_factory(type_info)
                if maybe_spec is not None:
                    return maybe_spec

        return UnsupportedTypeAnnotationError(
            (
                fmt.text(
                    "Unsupported type annotation ",
                    fmt.text["cyan"](str(type_info.type)),
                ),
            )
        )

    @classmethod
    def get_struct_spec(
        cls, type_info: StructTypeInfo
    ) -> StructConstructorSpec | InvalidDefaultInstanceError | None:
        """Get a constructor specification for a given type.

        Returns:
            - StructConstructorSpec if the type can be handled as a struct.
            - InvalidDefaultInstanceError if the type can be handled, but the provided default instance is incompatible with the type.
            - None if the type cannot be handled as a struct (no matching rule found).
        """

        cls._ensure_defaults_initialized()

        if (
            check_default_instances()
            and not is_sentinel(type_info.default)
            and not _resolver.is_instance(type_info.type, type_info.default)
        ):
            return InvalidDefaultInstanceError(
                (
                    fmt.text(
                        "Default value ",
                        fmt.text["cyan"](repr(type_info.default)),
                        " with type ",
                        fmt.text["magenta"](type(type_info.default).__name__),
                        " does not match type ",
                        fmt.text["magenta"](str(type_info.type)),
                    ),
                )
            )

        with type_info._typevar_context:
            for registry in cls._active_registries[::-1]:
                for spec_factory in registry._struct_rules[::-1]:
                    maybe_spec = spec_factory(type_info)
                    if maybe_spec is not None:
                        return maybe_spec

        return None

    def __enter__(self) -> None:
        self.__class__._ensure_defaults_initialized()

        cls = self.__class__
        cls._active_registries.append(self)

    def __exit__(self, *args: Any) -> None:
        cls = self.__class__
        assert cls._active_registries.pop() is self

    @classmethod
    def _ensure_defaults_initialized(cls) -> None:
        """Initialize default registry if needed."""
        if len(cls._active_registries) == 0:
            registry = ConstructorRegistry()

            # Apply the default rules.
            apply_default_primitive_rules(registry)
            apply_default_struct_rules(registry)

            cls._active_registries.append(registry)



"""Tests for collection[Struct] in unions creating proper subcommands.

This verifies that when collections (dict, list, tuple) have non-primitive values
and appear in a union, tyro correctly recognizes them as struct types and creates
subcommands.

Important: Collection types in unions require explicit defaults. Without a default,
even in a union context, they will error. This is consistent with tyro's philosophy
that standalone List[Struct], Dict[str, Struct], etc. require defaults.

Examples:
- List[Struct] | None           → ERROR (no default for list variant)
- List[Struct] | None = None    → ERROR (None doesn't provide default for list)
- List[Struct] | None = []      → OK (explicit empty list default)
- List[Struct] | None = [...]   → OK (explicit list default)
"""

import dataclasses
from typing import Dict, List, Optional, Set, Tuple, Union

import pytest

import tyro


def test_dict_struct_union_with_none() -> None:
    """Test that dict[str, Struct] | None creates subcommands."""

    @dataclasses.dataclass
    class Config:
        x: int
        y: int

    @dataclasses.dataclass
    class Args:
        # Creates subcommands for dict vs None.
        configs: Optional[Dict[str, Config]] = dataclasses.field(
            default_factory=lambda: {
                "a": Config(x=1, y=2),
                "b": Config(x=3, y=4),
            }
        )

    # Test default case (should use the dict variant).
    result = tyro.cli(Args, args=[])
    assert result.configs == {"a": Config(x=1, y=2), "b": Config(x=3, y=4)}

    # Test explicit dict subcommand with modified values.
    result = tyro.cli(
        Args,
        args=[
            "configs:dict-str-config",
            "--configs.a.x",
            "10",
            "--configs.b.y",
            "20",
        ],
    )
    assert result.configs == {"a": Config(x=10, y=2), "b": Config(x=3, y=20)}

    # Test None subcommand.
    result = tyro.cli(Args, args=["configs:none"])
    assert result.configs is None


def test_dict_struct_empty_default_union() -> None:
    """Test that dict[str, Struct] | None with empty default creates subcommands."""

    @dataclasses.dataclass
    class Config:
        x: int
        y: str

    @dataclasses.dataclass
    class Args:
        # Empty default - should still recognize as struct type.
        configs: Optional[Dict[str, Config]] = dataclasses.field(default_factory=dict)

    # Test None subcommand.
    result = tyro.cli(Args, args=["configs:none"])
    assert result.configs is None

    # With empty default dict provided.
    result = tyro.cli(Args, args=["configs:dict-str-config"])
    assert result.configs == {}


def test_dict_struct_union_with_other_type() -> None:
    """Test that dict[str, Struct] in union with other types creates subcommands."""

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class AlternativeConfig:
        name: str

    @dataclasses.dataclass
    class Args:
        data: Union[Dict[str, Config], AlternativeConfig] = dataclasses.field(
            default_factory=lambda: {"default": Config(value=5)}
        )

    # Test default (dict variant).
    result = tyro.cli(Args, args=[])
    assert result.data == {"default": Config(value=5)}

    # Test dict subcommand.
    result = tyro.cli(Args, args=["data:dict-str-config", "--data.default.value", "10"])
    assert result.data == {"default": Config(value=10)}

    # Test alternative config subcommand.
    result = tyro.cli(Args, args=["data:alternative-config", "--data.name", "test"])
    assert result.data == AlternativeConfig(name="test")


def test_nested_dict_struct() -> None:
    """Test that regular dict[str, Struct] (without union) still unpacks normally."""

    @dataclasses.dataclass
    class Config:
        x: int
        y: int

    @dataclasses.dataclass
    class Args:
        # Without union, should unpack the dict fields normally.
        configs: Dict[str, Config] = dataclasses.field(
            default_factory=lambda: {
                "a": Config(x=1, y=2),
                "b": Config(x=3, y=4),
            }
        )

    # Test that nested fields are accessible directly.
    result = tyro.cli(
        Args,
        args=["--configs.a.x", "10", "--configs.b.y", "20"],
    )
    assert result.configs == {"a": Config(x=10, y=2), "b": Config(x=3, y=20)}

    # Test default.
    result = tyro.cli(Args, args=[])
    assert result.configs == {"a": Config(x=1, y=2), "b": Config(x=3, y=4)}


def test_list_struct_union_with_other_type2() -> None:
    """Test that list[Struct] with non-empty default in union with struct works."""

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class AlternativeConfig:
        name: str

    @dataclasses.dataclass
    class Args:
        # Non-empty list default in union with another struct type.
        data: Union[List[Config], AlternativeConfig] = dataclasses.field(
            default_factory=lambda: [Config(value=5), Config(value=10)]
        )

    # Test default (list variant).
    result = tyro.cli(Args, args=[])
    assert result.data == [Config(value=5), Config(value=10)]

    # Test list subcommand.
    result = tyro.cli(Args, args=["data:list-config", "--data.0.value", "100"])
    assert result.data == [Config(value=100), Config(value=10)]

    # Test alternative config subcommand.
    result = tyro.cli(Args, args=["data:alternative-config", "--data.name", "test2"])
    assert result.data == AlternativeConfig(name="test2")


def test_list_struct_union_with_none_nonempty_default() -> None:
    """Test that list[Struct] | None with non-empty default creates subcommands."""

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class Args:
        # Non-empty default in union with None.
        data: Optional[List[Config]] = dataclasses.field(
            default_factory=lambda: [Config(value=5), Config(value=10)]
        )

    # Test default (list variant).
    result = tyro.cli(Args, args=[])
    assert result.data == [Config(value=5), Config(value=10)]

    # Test list subcommand with modified values.
    result = tyro.cli(Args, args=["data:list-config", "--data.0.value", "100"])
    assert result.data == [Config(value=100), Config(value=10)]

    # Test None subcommand.
    result = tyro.cli(Args, args=["data:none"])
    assert result.data is None


def test_list_struct_union_with_none_empty_default() -> None:
    """Test that list[Struct] | None with empty default creates subcommands."""

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class Args:
        # Empty default in union with None.
        data: Optional[List[Config]] = dataclasses.field(default_factory=list)

    # Test None subcommand.
    result = tyro.cli(Args, args=["data:none"])
    assert result.data is None

    # Test list subcommand with empty list.
    result = tyro.cli(Args, args=["data:list-config"])
    assert result.data == []


def test_tuple_struct_union_with_other_type() -> None:
    """Test that tuple[Struct, ...] with non-empty default in union with struct works."""

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class AlternativeConfig:
        name: str

    @dataclasses.dataclass
    class Args:
        # Non-empty tuple default in union with another struct type.
        data: Union[Tuple[Config, ...], AlternativeConfig] = dataclasses.field(
            default_factory=lambda: (Config(value=5), Config(value=10))
        )

    # Test default (tuple variant).
    result = tyro.cli(Args, args=[])
    assert result.data == (Config(value=5), Config(value=10))

    # Test tuple subcommand.
    result = tyro.cli(
        Args, args=["data:tuple-config-ellipsis", "--data.0.value", "100"]
    )
    assert result.data == (Config(value=100), Config(value=10))

    # Test alternative config subcommand.
    result = tyro.cli(Args, args=["data:alternative-config", "--data.name", "test2"])
    assert result.data == AlternativeConfig(name="test2")


def test_tuple_struct_union_with_none_nonempty_default() -> None:
    """Test that tuple[Struct, ...] | None with non-empty default creates subcommands."""

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class Args:
        # Non-empty default in union with None.
        data: Optional[Tuple[Config, ...]] = dataclasses.field(
            default_factory=lambda: (Config(value=5), Config(value=10))
        )

    # Test default (tuple variant).
    result = tyro.cli(Args, args=[])
    assert result.data == (Config(value=5), Config(value=10))

    # Test tuple subcommand with modified values.
    result = tyro.cli(
        Args, args=["data:tuple-config-ellipsis", "--data.0.value", "100"]
    )
    assert result.data == (Config(value=100), Config(value=10))

    # Test None subcommand.
    result = tyro.cli(Args, args=["data:none"])
    assert result.data is None


def test_tuple_struct_union_with_none_empty_default() -> None:
    """Test that tuple[Struct, ...] | None with empty default creates subcommands."""

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class Args:
        # Empty default in union with None.
        data: Optional[Tuple[Config, ...]] = dataclasses.field(default_factory=tuple)

    # Test None subcommand.
    result = tyro.cli(Args, args=["data:none"])
    assert result.data is None

    # Test tuple subcommand with empty tuple.
    result = tyro.cli(Args, args=["data:tuple-config-ellipsis"])
    assert result.data == ()


def test_list_struct_union_with_other_type() -> None:
    """Test that list[Struct] in union with other types creates subcommands."""

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class AlternativeConfig:
        name: str

    @dataclasses.dataclass
    class Args:
        data: Union[List[Config], AlternativeConfig] = dataclasses.field(
            default_factory=lambda: [Config(value=5)]
        )

    # Test default (list variant).
    result = tyro.cli(Args, args=[])
    assert result.data == [Config(value=5)]

    # Test list subcommand.
    result = tyro.cli(Args, args=["data:list-config", "--data.0.value", "10"])
    assert result.data == [Config(value=10)]

    # Test alternative config subcommand.
    result = tyro.cli(Args, args=["data:alternative-config", "--data.name", "test"])
    assert result.data == AlternativeConfig(name="test")


def test_list_struct_union_no_default_errors() -> None:
    """Test that List[Struct] | None without default raises an error.

    This verifies that we require explicit defaults even in union contexts,
    maintaining consistency with standalone collection types.
    """

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class Args:
        # No default - should error!
        data: Optional[List[Config]]

    # Should raise SystemExit when trying to use the list variant.
    with pytest.raises(SystemExit):
        tyro.cli(Args, args=["data:list-config"])


def test_dict_struct_union_no_default_errors() -> None:
    """Test that Dict[str, Struct] | None without default raises an error."""

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class Args:
        # No default - should error!
        data: Optional[Dict[str, Config]]

    # Should raise SystemExit when trying to use the dict variant.
    with pytest.raises(SystemExit):
        tyro.cli(Args, args=["data:dict-str-config"])


def test_tuple_struct_union_no_default_errors() -> None:
    """Test that Tuple[Struct, ...] | None without default raises an error."""

    @dataclasses.dataclass
    class Config:
        value: int

    @dataclasses.dataclass
    class Args:
        # No default - should error!
        data: Optional[Tuple[Config, ...]]

    # Should raise SystemExit when trying to use the tuple variant.
    with pytest.raises(SystemExit):
        tyro.cli(Args, args=["data:tuple-config-ellipsis"])


def test_set_struct_union_with_other_type() -> None:
    """Test that set[Struct] with non-empty default in union with struct works."""

    @dataclasses.dataclass(frozen=True)  # frozen=True for hashability.
    class Config:
        value: int

    @dataclasses.dataclass
    class AlternativeConfig:
        name: str

    @dataclasses.dataclass
    class Args:
        # Non-empty set default in union with another struct type.
        data: Union[Set[Config], AlternativeConfig] = dataclasses.field(
            default_factory=lambda: {Config(value=5), Config(value=10)}
        )

    # Test default (set variant).
    result = tyro.cli(Args, args=[])
    assert result.data == {Config(value=5), Config(value=10)}

    # Test alternative config subcommand.
    result = tyro.cli(Args, args=["data:alternative-config", "--data.name", "test2"])
    assert result.data == AlternativeConfig(name="test2")


def test_set_struct_union_with_none_nonempty_default() -> None:
    """Test that set[Struct] | None with non-empty default creates subcommands."""

    @dataclasses.dataclass(frozen=True)  # frozen=True for hashability.
    class Config:
        value: int

    @dataclasses.dataclass
    class Args:
        # Non-empty default in union with None.
        data: Optional[Set[Config]] = dataclasses.field(
            default_factory=lambda: {Config(value=5), Config(value=10)}
        )

    # Test default (set variant).
    result = tyro.cli(Args, args=[])
    assert result.data == {Config(value=5), Config(value=10)}

    # Test None subcommand.
    result = tyro.cli(Args, args=["data:none"])
    assert result.data is None


def test_set_struct_union_with_none_empty_default() -> None:
    """Test that set[Struct] | None with empty default creates subcommands.

    This test specifically covers the code path on line 405-406 in
    src/tyro/constructors/_struct_spec.py that handles empty set defaults.
    """

    @dataclasses.dataclass(frozen=True)  # frozen=True for hashability.
    class Config:
        value: int

    @dataclasses.dataclass
    class Args:
        # Empty default in union with None.
        data: Optional[Set[Config]] = dataclasses.field(default_factory=set)

    # Test None subcommand.
    result = tyro.cli(Args, args=["data:none"])
    assert result.data is None

    # Test set subcommand with empty set.
    result = tyro.cli(Args, args=["data:set-config"])
    assert result.data == set()


def test_set_struct_union_no_default_errors() -> None:
    """Test that set[Struct] | None without default raises an error."""

    @dataclasses.dataclass(frozen=True)  # frozen=True for hashability.
    class Config:
        value: int

    @dataclasses.dataclass
    class Args:
        # No default - should error!
        data: Optional[Set[Config]]

    # Should raise SystemExit when trying to use the set variant.
    with pytest.raises(SystemExit):
        tyro.cli(Args, args=["data:set-config"])



"""Core public API."""

from __future__ import annotations

import pathlib
import shutil
import sys
import warnings
from typing import Callable, Literal, Sequence, TypeVar, cast, overload

from typing_extensions import Annotated, assert_never, deprecated

from . import (
    _arguments,
    _calling,
    _parsers,
    _resolver,
    _settings,
    _strings,
    _unsafe_cache,
    conf,
)
from . import _fmtlib as fmt
from ._backends import _argparse as argparse
from ._singleton import (
    MISSING_NONPROP,
    NonpropagatingMissingType,
    PropagatingMissingType,
)
from ._typing import TypeForm
from .constructors import ConstructorRegistry
from .constructors._primitive_spec import UnsupportedTypeAnnotationError

OutT = TypeVar("OutT")


# The overload here is necessary for pyright and pylance due to special-casing
# related to using typing.Type[] as a temporary replacement for
# typing.TypeForm[].
#
# https://github.com/microsoft/pyright/issues/4298


@overload
def cli(
    f: TypeForm[OutT],
    *,
    prog: None | str = None,
    description: None | str = None,
    args: None | Sequence[str] = None,
    default: OutT
    | NonpropagatingMissingType
    | PropagatingMissingType = MISSING_NONPROP,
    return_unknown_args: Literal[False] = False,
    use_underscores: bool = False,
    console_outputs: bool = True,
    add_help: bool = True,
    compact_help: bool = False,
    config: None | Sequence[conf._markers.Marker] = None,
    registry: None | ConstructorRegistry = None,
) -> OutT: ...


@overload
def cli(
    f: TypeForm[OutT],
    *,
    prog: None | str = None,
    description: None | str = None,
    args: None | Sequence[str] = None,
    default: OutT
    | NonpropagatingMissingType
    | PropagatingMissingType = MISSING_NONPROP,
    return_unknown_args: Literal[True],
    use_underscores: bool = False,
    console_outputs: bool = True,
    add_help: bool = True,
    compact_help: bool = False,
    config: None | Sequence[conf._markers.Marker] = None,
    registry: None | ConstructorRegistry = None,
) -> tuple[OutT, list[str]]: ...


@overload
def cli(
    f: Callable[..., OutT],
    *,
    prog: None | str = None,
    description: None | str = None,
    args: None | Sequence[str] = None,
    # Passing a default makes sense for things like dataclasses, but are not
    # supported for general callables. These can, however, be specified in the
    # signature of the callable itself.
    default: NonpropagatingMissingType | PropagatingMissingType = MISSING_NONPROP,
    return_unknown_args: Literal[False] = False,
    use_underscores: bool = False,
    console_outputs: bool = True,
    add_help: bool = True,
    compact_help: bool = False,
    config: None | Sequence[conf._markers.Marker] = None,
    registry: None | ConstructorRegistry = None,
) -> OutT: ...


@overload
def cli(
    f: Callable[..., OutT],
    *,
    prog: None | str = None,
    description: None | str = None,
    args: None | Sequence[str] = None,
    # Passing a default makes sense for things like dataclasses, but are not
    # supported for general callables. These can, however, be specified in the
    # signature of the callable itself.
    default: NonpropagatingMissingType | PropagatingMissingType = MISSING_NONPROP,
    return_unknown_args: Literal[True],
    use_underscores: bool = False,
    console_outputs: bool = True,
    add_help: bool = True,
    compact_help: bool = False,
    config: None | Sequence[conf._markers.Marker] = None,
    registry: None | ConstructorRegistry = None,
) -> tuple[OutT, list[str]]: ...


def cli(
    f: TypeForm[OutT] | Callable[..., OutT],
    *,
    prog: None | str = None,
    description: None | str = None,
    args: None | Sequence[str] = None,
    default: OutT
    | NonpropagatingMissingType
    | PropagatingMissingType = MISSING_NONPROP,
    return_unknown_args: bool = False,
    use_underscores: bool = False,
    console_outputs: bool = True,
    add_help: bool = True,
    compact_help: bool = False,
    config: None | Sequence[conf._markers.Marker] = None,
    registry: None | ConstructorRegistry = None,
    **deprecated_kwargs,
) -> OutT | tuple[OutT, list[str]]:
    """Generate a command-line interface from type annotations and populate the target with arguments.

    :func:`cli()` is the core function of tyro. It takes a type-annotated function or class
    and automatically generates a command-line interface to populate it from user arguments.

    Two main usage patterns are supported:

    1. With a function (CLI arguments become function parameters):

       .. code-block:: python

          import tyro

          def main(a: str, b: str) -> None:
              print(a, b)

          if __name__ == "__main__":
              tyro.cli(main)  # Parses CLI args, calls main() with them

    2. With a class (CLI arguments become object attributes):

       .. code-block:: python

          from dataclasses import dataclass
          from pathlib import Path

          import tyro

          @dataclass
          class Config:
              a: str
              b: str

          if __name__ == "__main__":
              config = tyro.cli(Config)  # Parses CLI args, returns populated AppConfig
              print(f"Config: {config}")

    Args:
        f: The function or type to populate from command-line arguments. This must have
            type-annotated inputs for tyro to work correctly.
        prog: The name of the program to display in the help text. If not specified, the
            script filename is used. This mirrors the argument from
            :py:class:`argparse.ArgumentParser()`.
        description: The description text shown at the top of the help output. If not
            specified, the docstring of `f` is used. This mirrors the argument from
            :py:class:`argparse.ArgumentParser()`.
        args: If provided, parse arguments from this sequence of strings instead of
            the command line. This is useful for testing or programmatic usage. This mirrors
            the argument from :py:meth:`argparse.ArgumentParser.parse_args()`.
        default: An instance to use for default values. This is only supported if ``f`` is a
            type like a dataclass or dictionary, not if ``f`` is a general
            callable like a function. This is useful for merging CLI arguments
            with values loaded from elsewhere, such as a config file. The
            default value is :data:`tyro.MISSING_NONPROP`.
        return_unknown_args: If True, returns a tuple of the output and a list of unknown
            arguments that weren't consumed by the parser. This mirrors the behavior of
            :py:meth:`argparse.ArgumentParser.parse_known_args()`.
        use_underscores: If True, uses underscores as word delimiters in the help text
            instead of hyphens. This only affects the displayed help; both underscores and
            hyphens are treated equivalently during parsing. The default (False) follows the
            GNU style guide for command-line options.
            https://www.gnu.org/software/libc/manual/html_node/Argument-Syntax.html
        console_outputs: If set to False, suppresses parsing errors and help messages.
            This is useful in distributed settings where tyro.cli() is called from multiple
            workers but console output is only desired from the main process.
        add_help: Add a -h/--help option to the parser. This mirrors the argument from
            :py:class:`argparse.ArgumentParser()`.
        compact_help: If True, use compact help format that omits full argument descriptions.
            This mode shows only ``--flag TYPE (default: value)`` instead of including
            the full docstring. When enabled, users can access full help with
            ``--help-verbose``. Only applies to the TyroBackend; ignored for ArgparseBackend.
        config: A sequence of configuration marker objects from :mod:`tyro.conf`. This
            allows applying markers globally instead of annotating individual fields.
            For example: ``tyro.cli(Config, config=(tyro.conf.PositionalRequiredArgs,))``
        registry: A :class:`tyro.constructors.ConstructorRegistry` instance containing custom
            constructor rules.

    Returns:
        If ``f`` is a type (like a dataclass), returns an instance of that type populated
        with values from the command line. If ``f`` is a function, calls the function with
        arguments from the command line and returns its result. If ``return_unknown_args``
        is True, returns a tuple of the result and a list of unused command-line arguments.
    """

    # Make sure we start on a clean slate. Some tests may fail without this due to
    # memory address conflicts.
    _unsafe_cache.clear_cache()

    try:
        with _strings.delimeter_context("_" if use_underscores else "-"):
            output = _cli_impl(
                f,
                prog=prog,
                description=description,
                args=args,
                default=default,
                return_parser=False,
                return_unknown_args=return_unknown_args,
                use_underscores=use_underscores,
                console_outputs=console_outputs,
                add_help=add_help,
                compact_help=compact_help,
                config=config,
                registry=registry,
                **deprecated_kwargs,
            )
    except UnsupportedTypeAnnotationError as e:
        # Format and display the error nicely.
        error_message = fmt.box["bright_red"](
            fmt.text["bright_red", "bold"]("Invalid input to tyro.cli()"),
            fmt.rows(
                fmt.text("Could not create CLI parser from the provided type."),
                fmt.hr["red"](),
                *[fmt.cols((fmt.text["dim"]("• "), 2), msg) for msg in e.message],
            ),
        )
        print(
            "\n".join(
                error_message.render(width=min(shutil.get_terminal_size()[0], 80))
            ),
            file=sys.stderr,
            flush=True,
        )
        sys.exit(2)

    # Prevent unnecessary memory usage.
    _unsafe_cache.clear_cache()

    if return_unknown_args:
        assert isinstance(output, tuple)
        run_with_args_from_cli = output[0]
        out = run_with_args_from_cli()
        while isinstance(out, _calling.DummyWrapper):
            out = out.__tyro_dummy_inner__
        return out, output[1]  # type: ignore
    else:
        run_with_args_from_cli = cast(Callable[[], OutT], output)
        out = run_with_args_from_cli()
        while isinstance(out, _calling.DummyWrapper):
            out = out.__tyro_dummy_inner__
        return out


@overload
@deprecated("get_parser() is deprecated and will be removed in a future version.")
def get_parser(
    f: TypeForm[OutT],
    *,
    prog: None | str = None,
    description: None | str = None,
    default: OutT
    | NonpropagatingMissingType
    | PropagatingMissingType = MISSING_NONPROP,
    use_underscores: bool = False,
    console_outputs: bool = True,
    add_help: bool = True,
    config: None | Sequence[conf._markers.Marker] = None,
    registry: None | ConstructorRegistry = None,
) -> argparse.ArgumentParser: ...


@overload
@deprecated("get_parser() is deprecated and will be removed in a future version.")
def get_parser(
    f: Callable[..., OutT],
    *,
    prog: None | str = None,
    description: None | str = None,
    default: OutT
    | NonpropagatingMissingType
    | PropagatingMissingType = MISSING_NONPROP,
    use_underscores: bool = False,
    console_outputs: bool = True,
    add_help: bool = True,
    config: None | Sequence[conf._markers.Marker] = None,
    registry: None | ConstructorRegistry = None,
) -> argparse.ArgumentParser: ...


@deprecated("get_parser() is deprecated and will be removed in a future version.")
def get_parser(
    f: TypeForm[OutT] | Callable[..., OutT],
    *,
    # We have no `args` argument, since this is only used when
    # parser.parse_args() is called.
    prog: None | str = None,
    description: None | str = None,
    default: OutT
    | NonpropagatingMissingType
    | PropagatingMissingType = MISSING_NONPROP,
    use_underscores: bool = False,
    console_outputs: bool = True,
    add_help: bool = True,
    config: None | Sequence[conf._markers.Marker] = None,
    registry: None | ConstructorRegistry = None,
) -> argparse.ArgumentParser:
    """Get an :py:class:`argparse.ArgumentParser` object that approximates the CLI generated
    by :func:`tyro.cli`. Useful for tools like ``sphinx-argparse``, ``argcomplete``, etc.

    .. deprecated:: 1.0.0

        This function is deprecated and will be removed in a future version.

    .. note::

        The returned parser uses argparse-style subparsers, which is less flexible than
        tyro's subcommand parser.

    For tab completion, we recommend using :func:`tyro.cli`'s built-in
    ``--tyro-write-completion`` flag.

    Args:
        f: The function or type to populate from command-line arguments.
        prog: The name of the program to display in the help text.
        description: The description text shown at the top of the help output.
        default: An instance to use for default values.
        use_underscores: If True, uses underscores as word delimiters in the help text.
        console_outputs: If set to False, suppresses parsing errors and help messages.
        add_help: Add a -h/--help option to the parser. This mirrors the argument from
            :py:class:`argparse.ArgumentParser()`.
        config: A sequence of configuration marker objects from :mod:`tyro.conf`.
        registry: A :class:`tyro.constructors.ConstructorRegistry` instance containing custom
            constructor rules.
    """
    with _strings.delimeter_context("_" if use_underscores else "-"):
        return cast(
            argparse.ArgumentParser,
            _cli_impl(
                f,
                prog=prog,
                description=description,
                args=None,
                default=default,
                return_parser=True,
                return_unknown_args=False,
                use_underscores=use_underscores,
                console_outputs=console_outputs,
                add_help=add_help,
                compact_help=False,  # get_parser() always uses verbose help.
                config=config,
                registry=registry,
            ),
        )


def _cli_impl(
    f: TypeForm[OutT] | Callable[..., OutT],
    *,
    prog: None | str = None,
    description: None | str,
    args: None | Sequence[str],
    default: OutT | NonpropagatingMissingType | PropagatingMissingType,
    return_parser: bool,
    return_unknown_args: bool,
    console_outputs: bool,
    add_help: bool,
    compact_help: bool,
    config: None | Sequence[conf._markers.Marker],
    registry: None | ConstructorRegistry = None,
    **deprecated_kwargs,
) -> (
    OutT
    | argparse.ArgumentParser
    | tuple[
        Callable[[], OutT],
        list[str],
    ]
):
    """Helper for stitching the `tyro` pipeline together."""

    if config is not None and len(config) > 0:
        f = Annotated[(f, *config)]  # type: ignore

    if "default_instance" in deprecated_kwargs:
        warnings.warn(
            "`default_instance=` is deprecated! use `default=` instead.", stacklevel=2
        )
        default = deprecated_kwargs["default_instance"]
    if deprecated_kwargs.get("avoid_subparsers", False):
        f = conf.AvoidSubcommands[f]  # type: ignore
        warnings.warn(
            "`avoid_subparsers=` is deprecated! use `tyro.conf.AvoidSubcommands[]`"
            " instead.",
            stacklevel=2,
        )

    # Resolve any aliases, apply custom constructors that are directly attached
    # to the input type, etc.
    f = _resolver.TypeParamResolver.resolve_params_and_aliases(f)

    # Internally, we distinguish between two concepts:
    # - "default", which is used for individual arguments.
    # - "default_instance", which is used for _fields_ (which may be broken down into
    #   one or many arguments, depending on various factors).
    #
    # This could be revisited.
    default_instance = default

    # Read and fix arguments. If the user passes in --field_name instead of
    # --field-name, correct for them.
    args = list(sys.argv[1:]) if args is None else list(args)

    # Fix arguments. This will modify all option-style arguments replacing
    # underscores with hyphens, or vice versa if use_underscores=True.
    # If two options are ambiguous, e.g., --a_b and --a-b, raise a runtime error.
    #
    # This is only done for the argparse backend; the tyro backend handles
    # conversion internally.
    modified_args: dict[str, str] | None = None
    backend_name = _settings._experimental_options["backend"]
    if backend_name == "argparse":
        modified_args = {}
        for index, arg in enumerate(args):
            if not arg.startswith("--"):
                continue

            if "=" in arg:
                argname, _, val = arg.partition("=")
                fixed = "--" + _strings.swap_delimeters(argname[2:]) + "=" + val
                del argname, val
            else:
                fixed = "--" + _strings.swap_delimeters(arg[2:])
            if (
                return_unknown_args
                and fixed in modified_args
                and modified_args[fixed] != arg
            ):
                raise RuntimeError(
                    "Ambiguous arguments: " + modified_args[fixed] + " and " + arg
                )
            modified_args[fixed] = arg
            args[index] = fixed

    # If we pass in the --tyro-print-completion or --tyro-write-completion flags: turn
    # formatting tags, and get the shell we want to generate a completion script for
    # (bash/zsh/tcsh).
    #
    # shtab also offers an add_argument_to() functions that fulfills a similar goal, but
    # manual parsing of argv is convenient for turning off formatting.
    #
    # Note: --tyro-print-completion is deprecated! --tyro-write-completion is less prone
    # to errors from accidental logging, print statements, etc.
    print_completion = False
    write_completion = False
    if len(args) >= 2:
        # We replace underscores with hyphens to accomodate for `use_undercores`.
        print_completion = args[0].replace("_", "-") == "--tyro-print-completion"
        write_completion = (
            len(args) >= 3 and args[0].replace("_", "-") == "--tyro-write-completion"
        )

    # Note: setting USE_RICH must happen before the parser specification is generated.
    # TODO: revisit this. Ideally we should be able to eliminate the global state
    # changes.
    completion_shell = None
    completion_target_path = None
    if print_completion or write_completion:
        completion_shell = args[1]
    if write_completion:
        completion_target_path = pathlib.Path(args[2])

    # Map a callable to the relevant CLI arguments + subparsers.
    with _settings.timing_context("Generate parser specification"):
        if registry is not None:
            with registry:
                parser_spec = _parsers.ParserSpecification.from_callable_or_type(
                    f,
                    markers=set(),
                    description=description,
                    parent_classes=set(),  # Used for recursive calls.
                    default_instance=default_instance,  # Overrides for default values.
                    intern_prefix="",  # Used for recursive calls.
                    extern_prefix="",  # Used for recursive calls.
                    subcommand_prefix="",
                    support_single_arg_types=False,
                    prog_suffix="",
                )
        else:
            parser_spec = _parsers.ParserSpecification.from_callable_or_type(
                f,
                markers=set(),
                description=description,
                parent_classes=set(),  # Used for recursive calls.
                default_instance=default_instance,  # Overrides for default values.
                intern_prefix="",  # Used for recursive calls.
                extern_prefix="",  # Used for recursive calls.
                subcommand_prefix="",
                support_single_arg_types=False,
                prog_suffix="",
            )

    # Initialize backend.
    if backend_name == "argparse":
        from ._backends._argparse_backend import ArgparseBackend

        backend = ArgparseBackend()
    elif backend_name == "tyro":
        from ._backends._tyro_backend import TyroBackend

        backend = TyroBackend()
    else:
        assert_never(backend_name)

    # Handle shell completion.
    if print_completion or write_completion:
        assert completion_shell in (
            "bash",
            "zsh",
            "tcsh",
        ), f"Shell should be one `bash`, `zsh`, or `tcsh`, but got {completion_shell}"

        # Determine program name for completion script.
        if prog is None:
            prog = sys.argv[0]

        # Sanitize prog for use in function/variable names by replacing
        # non-alphanumeric characters with underscores.
        safe_prog = "".join(c if c.isalnum() or c == "_" else "_" for c in prog)

        # Generate completion script using the backend's method.
        completion_script = backend.generate_completion(
            parser_spec,
            prog=prog,
            shell=completion_shell,  # type: ignore
            root_prefix=f"tyro_{safe_prog}",
        )

        if write_completion and completion_target_path != pathlib.Path("-"):
            assert completion_target_path is not None
            completion_target_path.write_text(completion_script)
        else:
            print(completion_script)
        sys.exit()

    # For backwards compatibility with get_parser().
    if return_parser:
        return backend.get_parser_for_completion(
            parser_spec, prog=prog, add_help=add_help
        )

    # Parse arguments using the backend.
    if prog is None:
        prog = sys.argv[0]

    with _settings.timing_context("Parsing arguments"):
        value_from_prefixed_field_name, unknown_args = backend.parse_args(
            parser_spec=parser_spec,
            args=args,
            prog=prog,
            return_unknown_args=return_unknown_args,
            console_outputs=console_outputs,
            add_help=add_help,
            compact_help=compact_help,
        )

    try:
        # Attempt to call `f` using whatever was passed in.
        get_out, consumed_keywords = _calling.callable_with_args(
            f,
            parser_spec,
            default_instance,
            value_from_prefixed_field_name,
            field_name_prefix="",
        )
    except _calling.InstantiationError as e:
        # Print prettier errors.
        # This doesn't catch errors raised directly by get_out(), since that's
        # called later! This is intentional, because we do less error handling
        # for the root callable. Relevant: the `field_name_prefix == ""`
        # condition in `callable_with_args()`!

        # Emulate argparse's error behavior when invalid arguments are passed in.
        error_box_rows: list[str | fmt.Element] = []
        if isinstance(e.arg, _arguments.ArgumentDefinition):
            display_name = (
                str(e.arg.lowered.metavar)
                if e.arg.is_positional()
                else "/".join(e.arg.lowered.name_or_flags)
            )
            error_box_rows.extend(
                [
                    fmt.text(
                        fmt.text["bright_red", "bold"](
                            f"Error parsing {display_name}:"
                        ),
                        " ",
                        e.message,
                    ),
                    fmt.hr["red"](),
                    "Argument helptext:",
                    fmt.cols(
                        ("", 4),
                        fmt.rows(
                            e.arg.get_invocation_text()[1],
                            _arguments.generate_argument_helptext(e.arg, e.arg.lowered),
                        ),
                    ),
                ]
            )
        else:
            error_box_rows.append(
                fmt.text(
                    fmt.text["bright_red", "bold"](
                        f"Error parsing {e.arg}:",
                    ),
                    " ",
                    e.message,
                ),
            )

        if add_help:
            error_box_rows.extend(
                [
                    fmt.hr["red"](),
                    fmt.text(
                        "For full helptext, see ",
                        fmt.text["bold"](f"{prog} --help"),
                    ),
                ]
            )
        print(
            fmt.box["red"](fmt.text["red"]("Value error"), fmt.rows(*error_box_rows)),
            file=sys.stderr,
            flush=True,
        )
        sys.exit(2)

    assert len(value_from_prefixed_field_name.keys() - consumed_keywords) == 0, (
        f"Parsed {value_from_prefixed_field_name.keys()}, but only consumed"
        f" {consumed_keywords}"
    )
    if return_unknown_args:
        assert unknown_args is not None, "Should have parsed with `parse_known_args()`"
        # If we're parsed unknown args, we should return the original args, not
        # the fixed ones.
        if modified_args is not None:
            unknown_args = [modified_args.get(arg, arg) for arg in unknown_args]
        return get_out, unknown_args  # type: ignore
    else:
        assert unknown_args is None, "Should have parsed with `parse_args()`"
        return get_out  # type: ignore



import dataclasses
from typing import TYPE_CHECKING, Optional, Union

from typing_extensions import Annotated

import tyro


@dataclasses.dataclass
class A:
    x: int


def test_union_from_mapping():
    base_configs = {
        "one": A(1),
        "two": A(2),
        "three": A(3),
    }
    ConfigUnion = tyro.extras.subcommand_type_from_defaults(base_configs)

    assert tyro.cli(ConfigUnion, args="one".split(" ")) == A(1)
    assert tyro.cli(ConfigUnion, args="two".split(" ")) == A(2)
    assert tyro.cli(ConfigUnion, args="two --x 4".split(" ")) == A(4)
    assert tyro.cli(ConfigUnion, args="three".split(" ")) == A(3)


def test_union_from_mapping_in_function():
    base_configs = {
        "one": A(1),
        "two": A(2),
        "three": A(3),
    }

    if TYPE_CHECKING:
        ConfigUnion = A
    else:
        ConfigUnion = tyro.extras.subcommand_type_from_defaults(base_configs)

    def main(config: ConfigUnion, flag: bool = False) -> Optional[A]:  # type: ignore
        if flag:
            return config
        return None

    assert tyro.cli(main, args="--flag config:one".split(" ")) == A(1)
    assert tyro.cli(main, args="--flag config:one --config.x 3".split(" ")) == A(3)
    assert tyro.cli(main, args="config:one --config.x 1".split(" ")) is None

    assert tyro.cli(main, args="--flag config:two".split(" ")) == A(2)
    assert tyro.cli(main, args="--flag config:two --config.x 3".split(" ")) == A(3)
    assert tyro.cli(main, args="config:two --config.x 1".split(" ")) is None


def test_union_from_mapping_with_none_default() -> None:
    """Test that default matching works correctly when None is one of the options.

    This is a regression test for a bug where the subcommand matcher would incorrectly
    match any value against None when None was the first option in the union.
    """

    @dataclasses.dataclass
    class Config:
        x: Annotated[
            Union[A, None],
            tyro.conf.arg(
                constructor=tyro.extras.subcommand_type_from_defaults(
                    {"none": None, "A": A(3)}
                )
            ),
        ] = None

    # When default=Config(x=A(5)), the default subcommand should be "x:A", not "x:none".
    result = tyro.cli(Config, default=Config(x=A(5)), args=[])
    assert result.x is not None, "Default should be A(5), not None"
    assert result.x.x == 5, f"Expected x=5, got x={result.x.x}"

    # Test that we can still explicitly select the none subcommand.
    result = tyro.cli(Config, default=Config(x=A(5)), args=["x:none"])
    assert result.x is None, "Should be None when x:none is selected"

    # Test that we can override the default value.
    result = tyro.cli(Config, default=Config(x=A(5)), args=["x:A", "--x.x", "10"])
    assert result.x is not None
    assert result.x.x == 10



from __future__ import annotations

import contextlib
import io
import pathlib
from typing import Generic, TypeVar, cast

import attr
import pytest
from attrs import define, field
from helptext_utils import get_helptext_with_checks

import tyro
import tyro._strings


def test_attrs_basic() -> None:
    @attr.s
    class ManyTypesA:
        i: tyro.conf.Positional[int] = attr.ib()
        s: str = attr.ib()
        f: float = attr.ib()
        p: pathlib.Path = attr.ib()
        ignored: int = attr.ib(default=3, init=False)

    # We can directly pass a dataclass to `tyro.cli()`:
    assert tyro.cli(
        ManyTypesA,
        args=[
            "5",
            "--s",
            "5",
            "--f",
            "5",
            "--p",
            "~",
        ],
    ) == ManyTypesA(i=5, s="5", f=5.0, p=pathlib.Path("~"))


def test_attrs_defaults() -> None:
    @attr.s
    class ManyTypesB:
        i: int = attr.ib()
        s: str = attr.ib()
        f: float = attr.ib(default=1.0)

    # We can directly pass a dataclass to `tyro.cli()`:
    assert tyro.cli(
        ManyTypesB,
        args=[
            "--i",
            "5",
            "--s",
            "5",
        ],
    ) == ManyTypesB(i=5, s="5", f=1.0)


def test_attrs_helptext() -> None:
    @attr.s
    class Helptext:
        """This docstring should be printed as a description."""

        x: int = attr.ib()  # Documentation 1

        # Documentation 2
        y: int = attr.ib()

        z: int = attr.ib(default=3)
        """Documentation 3"""

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(Helptext, args=["--help"])
    helptext = f.getvalue()
    assert cast(str, Helptext.__doc__) in helptext

    assert "Documentation 1" in helptext
    assert "Documentation 2" in helptext
    assert "Documentation 3" in helptext


def test_attrs_next_gen_and_factory() -> None:
    @define
    class Helptext:
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        z: int = field(factory=lambda: 3)
        """Documentation 3"""

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(Helptext, args=["--help"])
    helptext = f.getvalue()
    assert cast(str, Helptext.__doc__) in helptext
    assert "Documentation 1" in helptext
    assert "Documentation 2" in helptext
    assert "Documentation 3" in helptext


def test_attrs_default_instance() -> None:
    @attr.s
    class ManyTypesB:
        i: int = attr.ib()
        s: str = attr.ib()
        f: float = attr.ib(default=1.0)
        k: float = attr.ib(default=1.0)

    assert tyro.cli(
        ManyTypesB,
        args=[
            "--i",
            "5",
            "--s",
            "5",
        ],
        default=ManyTypesB(i=5, s="5", f=2.0),
    ) == ManyTypesB(i=5, s="5", f=2.0)
    assert tyro.cli(
        ManyTypesB,
        args=["--i", "5"],
        default=ManyTypesB(i=5, s="5", f=2.0),
    ) == ManyTypesB(i=5, s="5", f=2.0)


T = TypeVar("T")


def test_attrs_inheritance_with_same_typevar() -> None:
    @attr.s
    class A(Generic[T]):
        x: T = attr.ib()

    @attr.s
    class B(A[int], Generic[T]):
        y: T = attr.ib()

    assert "INT" in get_helptext_with_checks(B[int])
    assert "STR" not in get_helptext_with_checks(B[int])
    assert "STR" in get_helptext_with_checks(B[str])
    assert "INT" in get_helptext_with_checks(B[str])

    assert tyro.cli(B[str], args=["--x", "1", "--y", "2"]) == B(x=1, y="2")
    assert tyro.cli(B[int], args=["--x", "1", "--y", "2"]) == B(x=1, y=2)


def test_diamond_inheritance() -> None:
    @define(frozen=True)
    class DiamondA:
        field: int | str = 5

    @define(frozen=True)
    class DiamondB(DiamondA):
        pass

    @define(frozen=True)
    class DiamondC(DiamondA):
        field: int = 10

    @define(frozen=True)
    class DiamondD(DiamondB, DiamondC):
        pass

    # C should come earlier int the MRO than A.
    helptext = get_helptext_with_checks(DiamondD)
    assert ": 5" not in helptext
    assert "10" in helptext
    assert "INT|STR" not in helptext


def test_attrs_init_false_with_default_instance() -> None:
    """Test that init=False fields preserve values from default instance (issue #390)."""

    @attr.define
    class AttrsConfig:
        in_channel: int = attr.field(init=False, default=0)
        out_channel: int = 3

    # Create a default instance and set the init=False field.
    default = AttrsConfig()
    default.in_channel = 10

    # The init=False field should preserve its value from the default instance.
    config = tyro.cli(AttrsConfig, default=default, args=["--out-channel", "5"])
    assert config.in_channel == 10
    assert config.out_channel == 5

    # Test with no args (should use default values).
    config2 = tyro.cli(AttrsConfig, default=default, args=[])
    assert config2.in_channel == 10
    assert config2.out_channel == 3

    # Test with falsy value (0) to ensure we don't skip None check incorrectly.
    default_falsy = AttrsConfig()
    default_falsy.in_channel = 0
    config3 = tyro.cli(AttrsConfig, default=default_falsy, args=["--out-channel", "7"])
    assert config3.in_channel == 0
    assert config3.out_channel == 7


def test_attrs_init_false_with_frozen() -> None:
    """Test that init=False fields work with frozen attrs classes (issue #415)."""

    @define(frozen=True)
    class FrozenAttrsConfig:
        in_channel: int = field(init=False, default=5)
        out_channel: int = 3

    default_config = FrozenAttrsConfig()

    result = tyro.cli(
        FrozenAttrsConfig, default=default_config, args=["--out-channel", "7"]
    )
    assert result.in_channel == 5
    assert result.out_channel == 7



"""ML collections

:func:`tyro.cli` understands and can populate config objects implemented using
`ml_collections <https://github.com/google/ml_collections/>`_, which is an
excellent library from folks at Google.

``ml_collections`` structures aren't statically typed, so we infer field types
based on value.

Usage:

    python ./08_ml_collections.py --help
    python ./08_ml_collections.py --config.network.policy-layer-dims 64 64 64
"""

from pprint import pprint

from ml_collections import ConfigDict, FrozenConfigDict  # type: ignore

import tyro


def get_config() -> FrozenConfigDict:
    config = ConfigDict()

    # Wandb config.
    config.wandb = ConfigDict()
    config.wandb.mode = "online"  # online, offline, disabled.
    config.wandb.project = "robot-sandbox"

    # Network config.
    config.network = ConfigDict()
    config.network.policy_layer_dims = (128,) * 3
    config.network.value_layer_dims = (256,) * 5
    config.network.policy_obs_key = "state"
    config.network.value_obs_key = "state"

    return FrozenConfigDict(config)


def train(config: FrozenConfigDict = get_config()) -> None:
    """Train a model."""
    pprint(config.to_dict())  # type: ignore


if __name__ == "__main__":
    tyro.cli(train)



from __future__ import annotations

from pydantic.dataclasses import dataclass
from typing_extensions import Annotated

import tyro


@dataclass
class Test:
    x: Annotated[str | int, "A string or an integer"]


def test_simple() -> None:
    assert tyro.cli(Test, args=["--x", "hello"]) == Test(x="hello")



import contextlib
import copy
import dataclasses
import io
import pathlib
from collections import namedtuple
from typing import Any, Dict, Mapping, NamedTuple, Tuple, Union, cast

import pytest
from typing_extensions import Literal, NotRequired, Required, TypedDict

import tyro
import tyro._strings


def test_basic_dict() -> None:
    def main(params: Dict[str, int]) -> Dict[str, int]:
        return params

    assert tyro.cli(main, args="--params hey 5 hello 2".split(" ")) == {
        "hey": 5,
        "hello": 2,
    }
    assert tyro.cli(main, args="--params hey 5 hello 2".split(" ")) == {
        "hey": 5,
        "hello": 2,
    }
    assert tyro.cli(main, args="--params".split(" ")) == {}
    with pytest.raises(SystemExit):
        tyro.cli(main, args="--params hey 5 hello hey".split(" "))
    with pytest.raises(SystemExit):
        tyro.cli(main, args="--params hey 5 hello".split(" "))


def test_dict_with_default() -> None:
    def main(params: Mapping[Literal[1, 3, 5, 7], bool] = {5: False, 1: True}) -> Any:
        return params

    assert tyro.cli(main, args=[]) == {5: False, 1: True}
    assert tyro.cli(main, args="--params.5 --params.no-1".split(" ")) == {
        5: True,
        1: False,
    }
    with pytest.raises(SystemExit):
        tyro.cli(main, args="--params".split(" "))


def test_tuple_in_dict() -> None:
    def main(x: Dict[Union[Tuple[int, int], Tuple[str, str]], Tuple[int, int]]) -> dict:
        return x

    assert tyro.cli(main, args="--x 1 1 2 2 3 3 4 4".split(" ")) == {
        (1, 1): (2, 2),
        (3, 3): (4, 4),
    }


def test_basic_typeddict() -> None:
    class ManyTypesTypedDict(TypedDict):
        i: int
        s: str

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--i 5 --s 5".split(" "),
    ) == dict(i=5, s="5")

    with pytest.raises(SystemExit):
        tyro.cli(ManyTypesTypedDict, args="--i 5".split(" "))

    with pytest.raises(SystemExit):
        tyro.cli(ManyTypesTypedDict, args="--s 5".split(" "))


def test_positional_in_typeddict() -> None:
    class ManyTypesTypedDict(TypedDict):
        i: tyro.conf.Positional[int]
        s: str

    assert tyro.cli(
        ManyTypesTypedDict,
        args="5 --s 5".split(" "),
    ) == dict(i=5, s="5")

    with pytest.raises(SystemExit):
        tyro.cli(ManyTypesTypedDict, args="5".split(" "))

    with pytest.raises(SystemExit):
        tyro.cli(ManyTypesTypedDict, args="--s 5".split(" "))


def test_total_false_typeddict() -> None:
    class ManyTypesTypedDict(TypedDict, total=False):
        i: int
        s: str

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--i 5 --s 5".split(" "),
    ) == dict(i=5, s="5")

    assert tyro.cli(ManyTypesTypedDict, args=[]) == dict()
    assert tyro.cli(ManyTypesTypedDict, args="--i 5".split(" ")) == dict(i=5)
    assert tyro.cli(ManyTypesTypedDict, args="--s 5".split(" ")) == dict(s="5")


def test_total_false_required_typeddict() -> None:
    class ManyTypesTypedDict(TypedDict, total=False):
        i: int
        s: Required[str]

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--i 5 --s 5".split(" "),
    ) == dict(i=5, s="5")

    with pytest.raises(SystemExit):
        # `s` is Required[].
        assert tyro.cli(ManyTypesTypedDict, args="--i 5".split(" ")) == dict(i=5)
    assert tyro.cli(
        ManyTypesTypedDict, args="--i 5".split(" "), default={"s": "5"}
    ) == dict(i=5, s="5")
    assert tyro.cli(ManyTypesTypedDict, args="--s 5".split(" ")) == dict(s="5")


def test_total_true_not_required_typeddict() -> None:
    class ManyTypesTypedDict(TypedDict, total=True):
        i: NotRequired[int]
        s: str

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--i 5 --s 5".split(" "),
    ) == dict(i=5, s="5")

    with pytest.raises(SystemExit):
        # `s` is Required[].
        assert tyro.cli(ManyTypesTypedDict, args="--i 5".split(" ")) == dict(i=5)
    assert tyro.cli(
        ManyTypesTypedDict, args="--i 5".split(" "), default={"s": "5"}
    ) == dict(i=5, s="5")
    assert tyro.cli(ManyTypesTypedDict, args="--s 5".split(" ")) == dict(s="5")


def test_total_false_nested_typeddict() -> None:
    class ChildTypedDict(TypedDict, total=False):
        i: int
        s: str

    class ParentTypedDict(TypedDict, total=False):
        child: ChildTypedDict

    assert tyro.cli(
        ParentTypedDict,
        args="--child.i 5 --child.s 5".split(" "),
    ) == {"child": {"i": 5, "s": "5"}}

    # total=False is ~ignored on the parent.
    assert tyro.cli(
        ParentTypedDict,
        args=[],
    ) == {"child": {}}


def test_total_false_typeddict_with_nested() -> None:
    @dataclasses.dataclass
    class Inner:
        j: float

    class ManyTypesTypedDict(TypedDict, total=False):
        i: int
        s: Inner

    # --s.j is (unfortunately) still required.
    with pytest.raises(SystemExit):
        tyro.cli(
            ManyTypesTypedDict,
            args="".split(" "),
        )

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--s.j 5".split(" "),
    ) == {"s": Inner(5.0)}


def test_total_false_typeddict_with_tuple() -> None:
    class ManyTypesTypedDict(TypedDict, total=False):
        i: int
        s: Tuple[str, str]

    assert (
        tyro.cli(
            ManyTypesTypedDict,
            args=[],
        )
        == dict()
    )

    assert tyro.cli(
        ManyTypesTypedDict,
        args="--i 5 --s 5 5".split(" "),
    ) == dict(i=5, s=("5", "5"))


def test_total_false_typeddict_with_union() -> None:
    """Test that total=False TypedDict with union field creates optional subparser.

    This tests the fix for EXCLUDE_FROM_CALL handling in union types.
    When a TypedDict field is optional and has a union type, EXCLUDE_FROM_CALL
    is used as the default, making the subparser optional. When no subcommand
    is selected, the field is excluded from the result.
    """

    @dataclasses.dataclass
    class StructA:
        a: int = 1

    @dataclasses.dataclass
    class StructB:
        b: int = 2

    class ConfigDict(TypedDict, total=False):
        choice: Union[StructA, StructB]
        other: int

    # Test 1: No arguments - should return empty dict (both fields excluded).
    assert tyro.cli(ConfigDict, args=[]) == {}

    # Test 2: Only 'other' argument - choice is excluded.
    assert tyro.cli(ConfigDict, args=["--other", "42"]) == {"other": 42}

    # Test 3: Only 'choice' subcommand - other is excluded.
    assert tyro.cli(ConfigDict, args=["choice:struct-a"]) == {"choice": StructA(a=1)}

    # Test 4: Both arguments.
    assert tyro.cli(ConfigDict, args=["--other", "99", "choice:struct-b"]) == {
        "choice": StructB(b=2),
        "other": 99,
    }


def test_not_required_with_union() -> None:
    """Test that NotRequired[] with union field creates optional subparser."""
    from typing_extensions import NotRequired

    @dataclasses.dataclass
    class StructA:
        a: int = 1

    @dataclasses.dataclass
    class StructB:
        b: int = 2

    class ConfigDict(TypedDict):
        choice: NotRequired[Union[StructA, StructB]]
        other: int

    # Test 1: Only required field.
    assert tyro.cli(ConfigDict, args=["--other", "42"]) == {"other": 42}

    # Test 2: Both fields.
    assert tyro.cli(ConfigDict, args=["--other", "99", "choice:struct-a"]) == {
        "choice": StructA(a=1),
        "other": 99,
    }


def test_required_vs_optional_union() -> None:
    """Test the difference between required and optional union subparsers."""

    @dataclasses.dataclass
    class StructA:
        a: int = 1

    @dataclasses.dataclass
    class StructB:
        b: int = 2

    # Required union (total=True by default).
    class RequiredConfig(TypedDict):
        choice: Union[StructA, StructB]

    # Optional union (total=False).
    class OptionalConfig(TypedDict, total=False):
        choice: Union[StructA, StructB]

    # Required: Must provide subcommand.
    with pytest.raises(SystemExit):
        tyro.cli(RequiredConfig, args=[])

    # Optional: Can omit subcommand.
    assert tyro.cli(OptionalConfig, args=[]) == {}


def test_nested_typeddict() -> None:
    class ChildTypedDict(TypedDict):
        y: int

    class NestedTypedDict(TypedDict):
        x: int
        b: ChildTypedDict

    assert tyro.cli(NestedTypedDict, args=["--x", "1", "--b.y", "3"]) == dict(
        x=1, b=dict(y=3)
    )
    with pytest.raises(SystemExit):
        tyro.cli(NestedTypedDict, args=["--x", "1"])


def test_helptext_and_default_typeddict() -> None:
    class HelptextTypedDict(TypedDict):
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        z: int
        """Documentation 3"""

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(HelptextTypedDict, default={"z": 3}, args=["--help"])
    helptext = tyro._strings.strip_ansi_sequences(f.getvalue())
    assert cast(str, HelptextTypedDict.__doc__) in helptext
    assert "--x INT" in helptext
    assert "--y INT" in helptext
    assert "--z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3 (default: 3)" in helptext


def test_basic_namedtuple() -> None:
    class ManyTypesNamedTuple(NamedTuple):
        i: int
        s: str
        f: float
        p: pathlib.Path

    assert tyro.cli(
        ManyTypesNamedTuple,
        args=[
            "--i",
            "5",
            "--s",
            "5",
            "--f",
            "5",
            "--p",
            "~",
        ],
    ) == ManyTypesNamedTuple(i=5, s="5", f=5.0, p=pathlib.Path("~"))


def test_collections_namedtuple_with_default() -> None:
    """Test that collections.namedtuple works with tyro.cli when default is provided."""
    SomeType = namedtuple("SomeType", ("field1", "field2", "field3"))

    # With a default value, tyro can infer types (int in this case)
    assert tyro.cli(
        SomeType,
        default=SomeType(0, 1, 2),
        args=["--field1", "3", "--field2", "4"],
    ) == SomeType(3, 4, 2)

    # Test with a mix of different types in default
    MixedType = namedtuple("MixedType", ("int_field", "str_field", "float_field"))
    assert tyro.cli(
        MixedType,
        default=MixedType(42, "hello", 3.14),
        args=["--int_field", "123", "--float_field", "2.718"],
    ) == MixedType(123, "hello", 2.718)


def test_collections_namedtuple_no_default_error() -> None:
    """Test that collections.namedtuple without default value raises the expected error."""
    SomeType = namedtuple("SomeType", ("field1", "field2", "field3"))

    # Without a default value, tyro can't infer types and should raise an error
    with pytest.raises(SystemExit):
        tyro.cli(
            SomeType,
            args=["--field1", "3", "--field2", "4", "--field3", "5"],
        )


def test_collections_namedtuple_with_defaults() -> None:
    """Test collections.namedtuple with _field_defaults dictionary."""
    # Create a namedtuple with defaults
    SomeTypeWithDefaults = namedtuple(
        "SomeTypeWithDefaults", ["field1", "field2", "field3"], defaults=(0, "default")
    )

    # The _field_defaults dict is automatically populated
    assert hasattr(SomeTypeWithDefaults, "_field_defaults")
    assert SomeTypeWithDefaults._field_defaults == {"field2": 0, "field3": "default"}

    # We need to provide a full instance as the default
    # The field defaults just populate _field_defaults dict but tyro still needs a default instance
    assert tyro.cli(
        SomeTypeWithDefaults,
        default=SomeTypeWithDefaults(5, 5, "default"),
        args=["--field1", "10", "--field2", "20"],
    ) == SomeTypeWithDefaults(10, 20, "default")


def test_nested_namedtuple() -> None:
    class ChildNamedTuple(NamedTuple):
        y: int

    class NestedNamedTuple(NamedTuple):
        x: int
        b: ChildNamedTuple

    assert tyro.cli(
        NestedNamedTuple, args=["--x", "1", "--b.y", "3"]
    ) == NestedNamedTuple(x=1, b=ChildNamedTuple(y=3))
    with pytest.raises(SystemExit):
        tyro.cli(NestedNamedTuple, args=["--x", "1"])


def test_helptext_and_default_namedtuple() -> None:
    class HelptextNamedTupleDefault(NamedTuple):
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        z: int = 3
        """Documentation 3"""

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(HelptextNamedTupleDefault, args=["--help"])
    helptext = tyro._strings.strip_ansi_sequences(f.getvalue())
    assert cast(str, HelptextNamedTupleDefault.__doc__) in helptext
    assert "--x INT" in helptext
    assert "--y INT" in helptext
    assert "--z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3 (default: 3)" in helptext


def test_helptext_and_default_namedtuple_alternate() -> None:
    class HelptextNamedTuple(NamedTuple):
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        z: int
        """Documentation 3"""

    with pytest.raises(SystemExit):
        tyro.cli(
            HelptextNamedTuple,
            default=tyro.MISSING,
            args=[],
        )

    f = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stdout(f):
            tyro.cli(
                HelptextNamedTuple,
                default=HelptextNamedTuple(
                    x=tyro.MISSING,
                    y=tyro.MISSING,
                    z=3,
                ),
                args=["--help"],
            )
    helptext = tyro._strings.strip_ansi_sequences(f.getvalue())
    assert cast(str, HelptextNamedTuple.__doc__) in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3" in helptext
    assert "(default: 3)" in helptext


def test_nested_dict() -> None:
    loaded_config = {
        "batch_size": 32,
        "optimizer": {
            "learning_rate": 1e-4,
            "epsilon": 1e-8,
            "scheduler": {"schedule_type": "constant"},
        },
    }
    backup_config = copy.deepcopy(loaded_config)
    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--batch-size",
            "16",
            "--optimizer.scheduler.schedule_type",
            "exponential",
        ],
    )

    # Overridden config should be different from loaded config.
    assert overrided_config != loaded_config
    assert overrided_config["batch_size"] == 16
    assert overrided_config["optimizer"]["scheduler"]["schedule_type"] == "exponential"

    # Original loaded config should not be mutated.
    assert loaded_config == backup_config


def test_nested_dict_use_underscores() -> None:
    loaded_config = {
        "batch_size": 32,
        "optimizer": {
            "learning_rate": 1e-4,
            "epsilon": 1e-8,
            "scheduler": {"schedule_type": "constant"},
        },
    }
    backup_config = copy.deepcopy(loaded_config)
    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--batch-size",
            "16",
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
        use_underscores=True,
    )

    # Overridden config should be different from loaded config.
    assert overrided_config != loaded_config
    assert overrided_config["batch_size"] == 16
    assert overrided_config["optimizer"]["scheduler"]["schedule_type"] == "exponential"

    # Original loaded config should not be mutated.
    assert loaded_config == backup_config


def test_nested_dict_hyphen() -> None:
    # We do a lot of underscore <=> conversion in the code; this is just to make sure it
    # doesn't break anything!
    loaded_config = {
        "batch-size": 32,
        "optimizer": {
            "learning-rate": 1e-4,
            "epsilon": 1e-8,
            "scheduler": {"schedule-type": "constant"},
        },
    }
    backup_config = copy.deepcopy(loaded_config)
    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--batch-size",
            "16",
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
    )

    # Overridden config should be different from loaded config.
    assert overrided_config != loaded_config
    assert overrided_config["batch-size"] == 16
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"

    # Original loaded config should not be mutated.
    assert loaded_config == backup_config


def test_nested_dict_hyphen_use_underscores() -> None:
    # We do a lot of underscore <=> conversion in the code; this is just to make sure it
    # doesn't break anything!
    loaded_config = {
        "batch-size": 32,
        "optimizer": {
            "learning-rate": 1e-4,
            "epsilon": 1e-8,
            "scheduler": {"schedule-type": "constant"},
        },
    }
    backup_config = copy.deepcopy(loaded_config)
    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--batch-size",
            "16",
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
        use_underscores=True,
    )

    # Overridden config should be different from loaded config.
    assert overrided_config != loaded_config
    assert overrided_config["batch-size"] == 16
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"

    # Original loaded config should not be mutated.
    assert loaded_config == backup_config

    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--batch_size",
            "16",
            "--optimizer.scheduler.schedule_type",
            "exponential",
        ],
        use_underscores=True,
    )

    # Overridden config should be different from loaded config.
    assert overrided_config != loaded_config
    assert overrided_config["batch-size"] == 16
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"

    # Original loaded config should not be mutated.
    assert loaded_config == backup_config


def test_nested_dict_annotations() -> None:
    loaded_config = {
        "optimizer": {
            "scheduler": {"schedule-type": "constant"},
        },
    }

    overrided_config = tyro.cli(
        dict,
        default=loaded_config,
        args=[
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
    )
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"
    del overrided_config

    overrided_config = tyro.cli(
        Dict[str, Dict],
        default=loaded_config,
        args=[
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
    )
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"
    del overrided_config

    overrided_config = tyro.cli(
        Dict[str, Dict[str, Dict]],
        default=loaded_config,
        args=[
            "--optimizer.scheduler.schedule-type",
            "exponential",
        ],
    )
    assert overrided_config["optimizer"]["scheduler"]["schedule-type"] == "exponential"
    del overrided_config


def test_functional_typeddict():
    """Source: https://github.com/brentyi/tyro/issues/87"""
    NerfMLPHiddenLayers_0 = TypedDict(
        "NerfMLPHiddenLayers_0",
        {
            "hidden_layers.0": int,
            "hidden_layers.1": int,
            "hidden_layers.2": int,
            "hidden_layers.3": int,
            "hidden_layers.4": int,
            "hidden_layers.5": int,
            "hidden_layers.6": int,
            "hidden_layers.7": int,
        },
    )
    NerfMLPHiddenLayers_1 = TypedDict(
        "NerfMLPHiddenLayers_1",
        {
            "hidden_layers.0": NotRequired[int],
            "hidden_layers.1": NotRequired[int],
            "hidden_layers.2": NotRequired[int],
            "hidden_layers.3": NotRequired[int],
            "hidden_layers.4": NotRequired[int],
            "hidden_layers.5": NotRequired[int],
            "hidden_layers.6": NotRequired[int],
            "hidden_layers.7": NotRequired[int],
        },
    )
    NerfMLPHiddenLayers_2 = TypedDict(
        "NerfMLPHiddenLayers_2",
        {
            "hidden_layers.0": int,
            "hidden_layers.1": int,
            "hidden_layers.2": int,
            "hidden_layers.3": int,
            "hidden_layers.4": int,
            "hidden_layers.5": int,
            "hidden_layers.6": int,
            "hidden_layers.7": int,
        },
        total=False,
    )
    with pytest.raises(SystemExit):
        tyro.cli(NerfMLPHiddenLayers_0, args=["--hidden_layers.0", "3"])
    assert tyro.cli(NerfMLPHiddenLayers_1, args=["--hidden_layers.0", "3"]) == {
        "hidden_layers.0": 3
    }
    assert tyro.cli(NerfMLPHiddenLayers_2, args=["--hidden_layers.0", "3"]) == {
        "hidden_layers.0": 3
    }


def test_not_required_bool() -> None:
    class NotRequiredBool(TypedDict):
        x: NotRequired[bool]

    assert tyro.cli(NotRequiredBool, args="--x".split(" ")) == {"x": True}
    assert tyro.cli(NotRequiredBool, args="--no-x".split(" ")) == {"x": False}
    assert tyro.cli(NotRequiredBool, args=[]) == {}


def test_functional_typeddict_with_default():
    """Source: https://github.com/brentyi/tyro/issues/87"""
    NerfMLPHiddenLayers_0 = TypedDict(
        "NerfMLPHiddenLayers_0",
        {
            "hidden_layers.0": int,
            "hidden_layers.1": int,
            "hidden_layers.2": int,
            "hidden_layers.3": int,
            "hidden_layers.4": int,
            "hidden_layers.5": int,
            "hidden_layers.6": int,
            "hidden_layers.7": int,
        },
    )
    NerfMLPHiddenLayers_1 = TypedDict(
        "NerfMLPHiddenLayers_1",
        {
            "hidden_layers.0": NotRequired[int],
            "hidden_layers.1": NotRequired[int],
            "hidden_layers.2": NotRequired[int],
            "hidden_layers.3": NotRequired[int],
            "hidden_layers.4": NotRequired[int],
            "hidden_layers.5": NotRequired[int],
            "hidden_layers.6": NotRequired[int],
            "hidden_layers.7": NotRequired[int],
        },
    )
    NerfMLPHiddenLayers_2 = TypedDict(
        "NerfMLPHiddenLayers_2",
        {
            "hidden_layers.0": int,
            "hidden_layers.1": int,
            "hidden_layers.2": int,
            "hidden_layers.3": int,
            "hidden_layers.4": int,
            "hidden_layers.5": int,
            "hidden_layers.6": int,
            "hidden_layers.7": int,
        },
        total=False,
    )
    with pytest.raises(SystemExit):
        tyro.cli(NerfMLPHiddenLayers_0, args=["--hidden_layers.0", "3"], default={})
    assert tyro.cli(
        NerfMLPHiddenLayers_1, args=["--hidden_layers.0", "3"], default={}
    ) == {"hidden_layers.0": 3}
    assert tyro.cli(
        NerfMLPHiddenLayers_2, args=["--hidden_layers.0", "3"], default={}
    ) == {"hidden_layers.0": 3}



import sys
from typing import Dict, FrozenSet, Generic, List, Literal, Optional, Set, TypeVar

import msgspec
import pytest
from helptext_utils import get_helptext_with_checks
from typing_extensions import NewType

import tyro

# Skip msgspec tests on Python 3.14 with msgspec <= 0.19.0
# See: https://github.com/jcrist/msgspec/issues/847
if sys.version_info >= (3, 14) and tuple(map(int, msgspec.__version__.split("."))) <= (
    0,
    19,
    0,
):
    pytest.skip(
        "msgspec incompatible with Python 3.14 (see msgspec issue #847)",
        allow_module_level=True,
    )

# Define TypeVars for generics tests
T = TypeVar("T")
S = TypeVar("S")
U = TypeVar("U", bound=float)

# Define NewType for testing
UserId = NewType("UserId", int)


def test_msgspec_with_complex_generics() -> None:
    """Test msgspec structs with complex generic parameters."""

    class Container(msgspec.Struct, Generic[T]):
        value: T
        name: str = "container"

    class NestedContainer(msgspec.Struct, Generic[T, S]):
        outer: Container[T]
        inner_value: S
        description: str = "nested"

    # Test with int, str combination
    result = tyro.cli(
        NestedContainer[int, str],
        args=[
            "--outer.value",
            "42",
            "--outer.name",
            "custom",
            "--inner-value",
            "hello",
        ],
    )
    assert result.outer.value == 42
    assert result.outer.name == "custom"
    assert result.inner_value == "hello"
    assert result.description == "nested"

    # Test with defaults
    result = tyro.cli(
        NestedContainer[float, bool],  # type: ignore[arg-type]
        args=["--outer.value", "3.14", "--inner-value", "True"],
    )
    assert result.outer.value == 3.14
    assert result.outer.name == "container"  # default
    assert result.inner_value is True
    assert result.description == "nested"  # default


def test_msgspec_with_bound_typevars() -> None:
    """Test msgspec structs with bounded TypeVars."""

    class NumericValue(msgspec.Struct, Generic[U]):
        value: U
        name: str = "numeric"

    # Test with int (within bounds)
    result = tyro.cli(NumericValue[int], args=["--value", "42"])
    assert result.value == 42.0
    assert result.name == "numeric"

    # Test with float (within bounds)
    result = tyro.cli(NumericValue[float], args=["--value", "3.14"])  # type: ignore[arg-type]
    assert result.value == 3.14
    assert result.name == "numeric"


def test_msgspec_with_generic_containers() -> None:
    """Test msgspec structs with generic container types."""

    class GenericListContainer(msgspec.Struct, Generic[T]):
        items: List[T]
        name: str = "list_container"

    # Test with list of strings
    result = tyro.cli(GenericListContainer[str], args=["--items", "a", "b", "c"])
    assert result.items == ["a", "b", "c"]
    assert result.name == "list_container"

    # Test with list of integers
    result = tyro.cli(GenericListContainer[int], args=["--items", "1", "2", "3"])  # type: ignore[arg-type]
    assert result.items == [1, 2, 3]
    assert result.name == "list_container"


def test_msgspec_with_generic_union_types() -> None:
    """Test msgspec structs with Union types inside generics."""

    class GenericOptional(msgspec.Struct, Generic[T]):
        value: Optional[T] = None
        description: str = "optional"

    # Test with None (default)
    result = tyro.cli(GenericOptional[int], args=[])
    assert result.value is None

    # Test with int value
    result = tyro.cli(GenericOptional[int], args=["--value", "42"])
    assert result.value == 42

    # Test with str value
    result = tyro.cli(GenericOptional[str], args=["--value", "hello"])  # type: ignore[arg-type]
    assert result.value == "hello"


def test_msgspec_with_literal_types() -> None:
    """Test msgspec structs with Literal types."""

    class ConfigMode(msgspec.Struct):
        mode: Literal["development", "production", "testing"] = "development"
        debug: bool = False

    # Test with default
    result = tyro.cli(ConfigMode, args=[])
    assert result.mode == "development"
    assert result.debug is False

    # Test with explicit mode
    result = tyro.cli(ConfigMode, args=["--mode", "production", "--debug"])
    assert result.mode == "production"
    assert result.debug is True

    # Test with invalid mode (should fail)
    with pytest.raises(SystemExit):
        tyro.cli(ConfigMode, args=["--mode", "invalid"])


def test_msgspec_with_newtype() -> None:
    """Test msgspec structs with NewType."""

    class User(msgspec.Struct):
        id: UserId
        name: str

    # Test with valid UserId
    result = tyro.cli(User, args=["--id", "12345", "--name", "John"])
    assert result.id == 12345
    assert isinstance(result.id, int)
    assert result.name == "John"


def test_msgspec_with_set_types() -> None:
    """Test msgspec structs with Set and FrozenSet types."""

    class SetContainer(msgspec.Struct):
        mutable_set: Set[int] = msgspec.field(default_factory=set)
        immutable_set: FrozenSet[str] = frozenset()

    # Test with default values
    result = tyro.cli(SetContainer, args=[])
    assert result.mutable_set == set()
    assert result.immutable_set == frozenset()

    # Test with custom values
    result = tyro.cli(
        SetContainer,
        args=["--mutable-set", "1", "2", "3", "--immutable-set", "a", "b", "c"],
    )
    assert result.mutable_set == {1, 2, 3}
    assert result.immutable_set == frozenset({"a", "b", "c"})


def test_msgspec_with_bytes() -> None:
    """Test msgspec structs with bytes and bytearray types."""

    class BinaryData(msgspec.Struct):
        data: bytes = b""

    # Test with default
    result = tyro.cli(BinaryData, args=[])
    assert result.data == b""

    # Test with string that gets converted to bytes
    # This relies on tyro's handling of bytes conversion
    result = tyro.cli(BinaryData, args=["--data", "hello"])
    assert result.data == b"hello"


def test_msgspec_with_dict() -> None:
    """Test msgspec structs with Dict type."""

    class DictContainer(msgspec.Struct):
        metadata: Dict[str, int] = msgspec.field(default_factory=dict)

    # Test with empty dict (default)
    result = tyro.cli(DictContainer, args=[])
    assert result.metadata == {}


def test_msgspec_with_generics_helptext() -> None:
    """Test helptext generation for generic msgspec structs."""

    class GenericConfig(msgspec.Struct, Generic[T]):
        """A generic configuration.

        This is used to configure generic values.
        """

        value: T
        """The generic value."""

        name: str = "config"
        """Configuration name."""

    # Verify helptext contains the docstrings
    helptext = get_helptext_with_checks(GenericConfig[int])
    assert "A generic configuration" in helptext
    assert "This is used to configure generic values" in helptext
    assert "The generic value" in helptext
    assert "Configuration name" in helptext

    # Verify the parameter type is shown correctly
    # For int instantiation
    helptext_int = get_helptext_with_checks(GenericConfig[int])
    assert "--value INT" in helptext_int

    # For str instantiation
    helptext_str = get_helptext_with_checks(GenericConfig[str])
    assert "--value STR" in helptext_str



from __future__ import annotations

from typing import Any, Mapping, Sequence, Tuple, TypeVar, Union

from typing_extensions import Annotated

from tyro.conf._markers import Suppress
from tyro.constructors import ConstructorRegistry

from .._typing import TypeForm

T = TypeVar("T")


def overridable_config_cli(
    configs: Mapping[str, Tuple[str, T]],
    *,
    prog: str | None = None,
    description: str | None = None,
    args: Sequence[str] | None = None,
    use_underscores: bool = False,
    console_outputs: bool = True,
    add_help: bool = True,
    config: Sequence[Any] | None = None,
    sort_subcommands: bool = False,
    registry: ConstructorRegistry | None = None,
) -> T:
    """Helper function for creating a CLI interface that allows us to choose
    between default config objects (typically dataclasses) and override values
    within it. Turns off subcommand creation for any union types within the
    config object.

    This is a lightweight wrapper over :func:`tyro.cli()`, with some default
    arguments populated. Also see
    :func:`tyro.extras.subcommand_type_from_defaults()`.


    Example usage:

    .. code-block:: python

        import dataclasses

        import tyro


        @dataclasses.dataclass
        class Config:
            a: int
            b: str


        default_configs = {
            "small": (
                "Small config",
                Config(1, "small"),
            ),
            "big": (
                "Big config",
                Config(100, "big"),
            ),
        }
        config = tyro.extras.overridable_config_cli(default_configs)
        print(config)

    Args:
        configs: A dictionary of config names mapped to a tuple of
            (description, config object).
        prog: The name of the program printed in helptext. Mirrors argument from
            `argparse.ArgumentParser()`.
        description: Description text for the parser, displayed when the --help flag is
            passed in. If not specified, the class docstring is used. Mirrors argument from
            `argparse.ArgumentParser()`.
        args: If set, parse arguments from a sequence of strings instead of the
            commandline. Mirrors argument from `argparse.ArgumentParser.parse_args()`.
        use_underscores: If True, use underscores as a word delimiter instead of hyphens.
            This primarily impacts helptext; underscores and hyphens are treated equivalently
            when parsing happens. We default helptext to hyphens to follow the GNU style guide.
            https://www.gnu.org/software/libc/manual/html_node/Argument-Syntax.html
        console_outputs: If set to `False`, parsing errors and help messages will be
            suppressed.
        add_help: Add a -h/--help option to the parser. This mirrors the argument from
            `argparse.ArgumentParser()`.
        config: Sequence of config marker objects, from `tyro.conf`. We include
            :class:`tyro.conf.AvoidSubcommands` by default.
        sort_subcommands: If True, sort the subcommands alphabetically by name.
        registry: A :class:`tyro.constructors.ConstructorRegistry` instance containing custom
            constructor rules.
    """
    import tyro

    keys = list(configs.keys())
    return tyro.cli(
        tyro.extras.subcommand_type_from_defaults(
            defaults={k: configs[k][1] for k in keys},
            descriptions={k: configs[k][0] for k in keys},
            sort_subcommands=sort_subcommands,
        ),
        prog=prog,
        description=description,
        args=args,
        use_underscores=use_underscores,
        console_outputs=console_outputs,
        add_help=add_help,
        # Don't create subcommands for union types within the config object.
        config=(tyro.conf.AvoidSubcommands,)
        + (tuple() if config is None else tuple(config)),
        registry=registry,
    )


def subcommand_type_from_defaults(
    defaults: Mapping[str, T],
    descriptions: Mapping[str, str] = {},
    *,
    prefix_names: bool = True,
    sort_subcommands: bool = False,
) -> TypeForm[T]:
    """Construct a Union type for defining subcommands that choose between defaults.

    For example, when ``defaults`` is set to:

    .. code-block:: python

        {
            "small": Config(...),
            "big": Config(...),
        }

    We return:

    .. code-block:: python

        Union[
            Annotated[
                Config,
                tyro.conf.subcommand("small", default=Config(...))
            ],
            Annotated[
                Config,
                tyro.conf.subcommand("big", default=Config(...))
            ]
        ]

    Direct use of :py:data:`typing.Union` and :func:`tyro.conf.subcommand()` should generally be
    preferred, but this function can be helpful for succinctness.

    .. warning::

        The type returned by this function can be safely used as an input to
        :func:`tyro.cli()`, but for static analysis when used for annotations we
        recommend applying a `TYPE_CHECKING` guard:

        .. code-block:: python

            from typing import TYPE_CHECKING

            if TYPE_CHECKING:
                # Static type seen by language servers, type checkers, etc.
                SelectableConfig = Config
            else:
                # Runtime type used by tyro.
                SelectableConfig = subcommand_type_from_defaults(...)

    Args:
        defaults: A dictionary of default subcommand instances.
        descriptions: A dictionary conttaining descriptions for helptext.
        prefix_names: Whether to prefix subcommand names.
        sort_subcommands: If True, sort the subcommands alphabetically by name.

    Returns:
        A subcommand type, which can be passed to :func:`tyro.cli`.
    """
    import tyro

    keys = list(defaults.keys())
    if sort_subcommands:
        keys = sorted(keys)
    return Union[  # type: ignore
        tuple(
            Annotated[  # type: ignore
                (
                    type(defaults[k]),
                    tyro.conf.subcommand(
                        k,
                        default=defaults[k],
                        description=descriptions.get(k, ""),
                        prefix_name=prefix_names,
                    ),
                )
            ]
            for k in keys
        )
        # Union types need at least two types. To support the case
        # where we only pass one subcommand in, we'll pad with `None`
        # but suppress it.
        + (Annotated[None, Suppress],)
    ]
